/* 
 * File:   CBlat.cpp
 * Author: mwittig
 * 
 * Created on January 17, 2013, 3:49 PM
 */

#include <string>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <sstream>
#include <map>
#include <fstream>
#include <vector>
#include <set>

#include "CBlat.h"

CBlat::CBlat()
{
    m_strExecutive = "blat";
    m_strError = "";
}

CBlat::CBlat(const string& executive)
{
    m_strExecutive = executive;
    m_strError = "";
}

CBlat::CBlat(const CBlat& orig)
{
    m_strExecutive              = orig.m_strExecutive;
    m_aligned_reads             = orig.m_aligned_reads;
    m_strError                  = orig.m_strError;
    m_paired_reads_detailed     = orig.m_paired_reads_detailed;
    m_unpaired_reads_detailed   = orig.m_unpaired_reads_detailed;
}

CBlat::~CBlat() {
}

int CBlat::cmd(const string& commando, bool write_output)
{
    FILE *in;
    char buff[512];

    if(!(in = popen(commando.c_str(), "r")))
    {
        return 1;
    }

    while(fgets(buff, sizeof(buff), in)!=NULL)
        if(write_output)
            cout << buff;
    pclose(in);
    if(write_output)
        cout << endl;
    return 0;
}

bool CBlat::run(const string& reference,const string& reads,const string& output, const long& read_length, const int tileSize)
{
    try
    {
        ostringstream strHlp("");

        strHlp << m_strExecutive << ' ' << reference << ' ' << reads << " -out=pslx -maxIntron=0 -minIdentity=100 -minScore=30 -fine -tileSize=" << tileSize << ' ' << output << endl;
        cmd(strHlp.str(),false);
        if(parseResults(output,read_length))
            return true;
        return false;
    }
    catch(...)
    {
        return false;
    }
}

bool CBlat::parseResults(const string& filename, const long& read_length)
{
    try
    {
        ifstream in(filename.c_str());
        if(!in)
        {
            cerr << "Can't open " << filename << endl;
            return false;
        }
        size_t truncated_reads = 0;
        string strLine;
        while(getline(in,strLine))
            if(strLine.length() > 0 && strLine[0] == '-')
                break;
        while(getline(in,strLine))
        {
            vector<string> parsed = Tokenize(strLine);
            //  0      1       2       3        4       5       6       7       8     9        10     11      12      13      14      15      16      17      18         19              20       21
            //match   mis-    rep.    N's     Q gap   Q gap   T gap   T gap   strand  Q        Q       Q       Q       T       T       T       T     block   blockSizes  qStarts     tStarts
            //match   match           count   bases   count   bases                  name    size    start   end     name    size    start   end     count
            stcAlignment actAlign;
            actAlign.RefId      = atol(parsed[13].c_str());
            actAlign.AlignLength = atol(parsed[0].c_str());
            actAlign.RefStart    = atol(parsed[15].c_str());
            actAlign.ReadStart   = atol(parsed[11].c_str());
            actAlign.strand      = parsed[8][0];
            if(actAlign.strand == '-')
            {
                size_t hlp = actAlign.ReadStart+actAlign.AlignLength;
                actAlign.ReadStart = read_length-hlp;
            }
            //cout << actAlign.RefId << '\t' << actAlign.RefStart << '\t' << actAlign.AlignLength << '\t' << actAlign.ReadStart << '\t' << actAlign.strand << endl;
            m_aligned_reads.insert(pair<string,stcAlignment>(parsed[9],actAlign));
            if(atol(parsed[10].c_str()) != read_length)
                truncated_reads++;
        }
        in.close();
        if(truncated_reads > 0)
            cerr << "found truncated reads: " << truncated_reads << "/" << m_aligned_reads.size() << " aligned reads." << endl;
        if((float)truncated_reads/m_aligned_reads.size() > 0.33)
            throw(truncated_reads);
        return true;
    }
    catch(size_t val)
    {
        cerr << "more than 33% of the reads are shorter or longer than the specified " << read_length << "bp. Program aborted ..." << endl;
        exit(EXIT_FAILURE);
    }
    catch(...)
    {
        return false;
    }
}

vector<string> CBlat::Tokenize(const string& str, const string& delimiters)
{
    vector<string> tokens;
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        if(str.length() > pos && delimiters.find(str[pos]) != string::npos )
            lastPos=pos+1;
        else
            lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
    return tokens;
}

void CBlat::storePairedReads(const string& separator)
{
    for(map<string,stcAlignment>::const_iterator iter = m_aligned_reads.begin(); iter != m_aligned_reads.end(); iter++)
    {
        string the_id = Tokenize(iter->first,"/")[0];
        map<string,stcAlignment>::iterator read_set_iter = m_unpaired_reads_detailed.find(the_id);
        if(read_set_iter != m_unpaired_reads_detailed.end())
        {
            m_paired_reads_detailed[the_id]=pair<stcAlignment,stcAlignment>(m_unpaired_reads_detailed[the_id],iter->second);
            m_unpaired_reads_detailed.erase(read_set_iter);
        }
        else
            m_unpaired_reads_detailed[the_id]=iter->second;
    }
}

void CBlat::removeTruncatedReads(const vector<string>& m_exon_sequences, const unsigned short read_length)
{
    // Aufzeichnen und neu machen
    // cout << m_aligned_reads.size() << " -> ";
    multimap<string,stcAlignment>::iterator iter = m_aligned_reads.begin();
    while(iter != m_aligned_reads.end())
    {
        stcAlignment& act_align = iter->second;
        unsigned int  exon_size = m_exon_sequences[act_align.RefId].size();
        unsigned int  qEnd      = act_align.ReadStart+act_align.AlignLength; // query end bedeutet Ende in der read sequenz
        unsigned int  tEnd      = act_align.RefStart+act_align.AlignLength; //  template end bedeutet Ende in der Referenzsequenz

        if(
             act_align.AlignLength == read_length || act_align.AlignLength == exon_size  || // der ganze read oder der read über das ganze exon
           ( act_align.RefStart  == 0 && qEnd == read_length && act_align.AlignLength >= MIN_TRUNCATED_READ_LENGTH ) ||  // der read hängt am ref Start & ist mindestens 70 bp
           ( act_align.ReadStart == 0 && tEnd == exon_size && act_align.AlignLength >= MIN_TRUNCATED_READ_LENGTH )   ||   // der read hängt am ref End & ist mindestens 70 bp
           ( act_align.ReadStart+act_align.AlignLength == read_length && act_align.RefStart+act_align.AlignLength == exon_size && act_align.strand == '-' && act_align.AlignLength >= MIN_TRUNCATED_READ_LENGTH ) ||
           ( act_align.ReadStart == 0 && act_align.AlignLength >= MIN_TRUNCATED_READ_LENGTH && act_align.RefStart  == 0 && act_align.strand == '-' )
          )
        {
            ++iter;
            continue; // So solls sein, gehe zum nächsten read
        }
        m_aligned_reads.erase(iter++); // Schmeiss den read raus
    }
    storePairedReads("/");
    // cout << m_aligned_reads.size() << " truncated" << endl;
}

void CBlat::removeRedundantAlignments()
{
    set<string> unifier;
    multimap<string,stcAlignment>::iterator iter = m_aligned_reads.begin();
    while( iter != m_aligned_reads.end())
    {
        ostringstream osr("");
        osr << iter->second.RefId << '_' << iter->second.ReadStart << '_' << iter->second.RefStart;
        if(unifier.find(osr.str()) == unifier.end())
        {
            ++iter;
            unifier.insert(osr.str());
        }
        else
            m_aligned_reads.erase(iter++);
    }

}


