/* 
 * File:   CHLAAlignment.h
 * Author: mwittig
 *
 * Created on January 24, 2013, 4:58 PM
 */

#ifndef _CHLAALIGNMENT_H
#define	_CHLAALIGNMENT_H

#include "CBlat.h"

//#define _CHLAALIGNMENT_READ_LENGTH 100
#define _CHLAALIGNMENT_MIN_ALIGNMENT 70
#define _CHLAALIGNMENT_MAX_SOFTCLIPPING (m_read_length-_CHLAALIGNMENT_MIN_ALIGNMENT)
#define _CHLAALIGNMENT_MIN_EXON_LENGTH 30
#define _CHLAALIGNMENT_MIN_MIN_CENTRAL_READS_SCORE 0.20f

using namespace std;

class CHLAAlignment {
public:

    class stcAlignment{
        public:
        stcAlignment(){ReadStart = ReadStart = AlignLength = 0; strand = '+';}
        stcAlignment(const stcAlignment& orig){
            ReadStart = orig.ReadStart;
            RefStart  = orig.RefStart;
            AlignLength= orig.AlignLength;
            strand    = orig.strand;
        }
        ~stcAlignment(){}

        unsigned int     RefStart;     // 0 basiert
        unsigned int     ReadStart;    // 0 basiert
        unsigned short   AlignLength;
        char             strand;
    };

    class stcExonCoverage{
        public:
            stcExonCoverage(){maxCoverage = startCoverage = endCoverage = 0; avgCoverage = medCoverage = 0.0f;m_central_reads_fraction = m_left_right_reads_fraction = vector<float>(1,1.0f);}
            stcExonCoverage(const stcExonCoverage& orig){
            avgCoverage     = orig.avgCoverage;
            maxCoverage     = orig.maxCoverage;
            medCoverage     = orig.medCoverage;
            startCoverage   = orig.startCoverage;
            endCoverage     = orig.endCoverage;
            m_central_reads_fraction = orig.m_central_reads_fraction;
            m_left_right_reads_fraction = orig.m_left_right_reads_fraction;
            }
            ~stcExonCoverage(){}

        // für die nächsten beiden vectoren is -1.0f das NA value
        vector<float>    m_central_reads_fraction;     // anteil an central reads an einer Position, ideal = 0.33333
        vector<float>    m_left_right_reads_fraction;  // mein erste central reads implementation, die eher eine links-rechts Verteilung ist

        double           avgCoverage;       // mean, average
        unsigned int     maxCoverage;       // Maximum
        double           medCoverage;       // Median
        unsigned int     startCoverage;     // Coverage am Start
        unsigned int     endCoverage;       // coverage am Ende
    };
    
    enum coverage_metrics {Median, Mean, Max};

    CHLAAlignment();
    CHLAAlignment(const string&); // Diesen Konstruktor nur verwenden, wenn nicht mit den reads gearbeitet wird
    CHLAAlignment(const string&, const unsigned int&);
    CHLAAlignment(const CHLAAlignment& orig);
    virtual ~CHLAAlignment();

    string getExonReport();

    string getExonsAsFasta();
    void   getExonsAsFasta(const string& filename);


    void transferReadsFromBlat(const multimap<string,CBlat::stcAlignment>&  aligned_reads);

    // für den Plot sind start und end 0-basiert. Zum rechnen intern besser mit 0-basiert start und 1-basiert end
    vector<pair<long,long> > getSortedReads(bool zero_based_end);
    void getSortedReads(const string& filename, bool);

    vector<long>             getCoverage();  // Hier sind die Koordiunaten 0-basiert
    void                     getCoverage(const string& filename);

    vector<float>            getCentralReadsScore();
    void                     getCentralReadsScore(const string& filename);

    vector<float>            getLeftRightScore();
    void                     getLeftRightScore(const string& filename);

    vector<long>             getExonBoundaries();  // Hier sind die Koordiunaten 0-basiert
    void                     getExonBoundaries(const string& filename);

    vector<long>             getIdealCoverage();
    void                     getIdealCoverage(const string& filename);

    vector<long>             getScoreFailed();
    void                     getScoreFailed(const string& filename);

    double                   getAreaUnderCurveError();

    size_t                   getNumberOfCoveredBases();
    size_t                   getSequenceSize();
    size_t                   getAlignableSequenceSize(size_t minLength = _CHLAALIGNMENT_MIN_EXON_LENGTH); // Exone < 30 bp werden als nicht alignbar gewertet
    string                   getContinuousSequence();

    size_t                   getExonBoundaryPenalties(coverage_metrics metrics = Mean ,const float factor = 0.33); // exon grenzen müssen Mindestens die factor * (Mean, Median oder Max überschreiten)
    void                     calculateCoverageMetrics();

    void                     calculateCoverageDropsExtended();
    vector<size_t>           findCoverageDropsExtended();
    vector<CHLAAlignment::stcAlignment>     getSortedMappings();

    string                   getAlignableSequence();
    long                     getAlignableExonCount();
    bool                     isAlignableExon(size_t nr){if(nr >= m_exon_sequences.size())return false; return m_exon_sequences.size() >= _CHLAALIGNMENT_MIN_EXON_LENGTH;};

    static unsigned int      getMinimumExonSize(){return _CHLAALIGNMENT_MIN_EXON_LENGTH;}

    vector<string>                              m_exon_sequences;
    vector<pair<unsigned int, unsigned int> >   m_exon_coordinates;

    map<string,CHLAAlignment::stcAlignment>     m_aligned_reads;

    vector<stcExonCoverage>                     m_coverage_metrics;

private:
    // Liefert die Start und Endkoordinaten der Exon Blöcke
    void getExonBlocks(const string& seq);
    static bool sort_alignments(const pair<long,long>&, const pair<long,long>&);
    static bool sort_mappings(const CHLAAlignment::stcAlignment& one, const CHLAAlignment::stcAlignment& two);

    vector<stcExonCoverage> getExonCoverageMetrics();
    double getMedianScore(vector<long> scores);

    pair<size_t,size_t>         countReadsAtAnchor(const vector<pair<long,long> >& ordered_reads,long anchor);
    vector<pair<long,long> >    getOverlappingReads(const vector<pair<long,long> >& ordered_reads,long pos);
    vector<CHLAAlignment::stcAlignment >       getOverlappingMappings(const vector<CHLAAlignment::stcAlignment>& ordered_mappings, pair<long,long> pos);
    double                      getCentralReadsRatio(vector<pair<long,long> > ordered_reads,long anchor);

    vector<double>              getAreaUnderCurveInterpolator();

    size_t                      getcDNAsize();

    unsigned int                m_read_length;

    bool doesAlignmentOverlapPosition(CHLAAlignment::stcAlignment,pair<long,long> pos);
    vector<string> Tokenize(const string& str, const string& delimiters = "\t");
};

#endif	/* _CHLAALIGNMENT_H */

