/* 
 * File:   CImgtAmbiguities.cpp
 * Author: mwittig
 * 
 * Created on March 13, 2013, 1:35 PM
 */
#include <string>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <sstream>
#include <map>
#include <fstream>
#include <vector>
#include <set>
#include <string.h>

#include "CImgtAmbiguities.h"


CImgtAmbiguities::CImgtAmbiguities() {
}

CImgtAmbiguities::CImgtAmbiguities(const string& filename_ambiguities,const string& filename_chimera)
{
    readAmbiguityTable(filename_ambiguities.c_str());
}

CImgtAmbiguities::CImgtAmbiguities(const CImgtAmbiguities& orig) {
}

CImgtAmbiguities::~CImgtAmbiguities() {
}

void CImgtAmbiguities::readAmbiguityTable(const char* filename)
{
    ifstream in(filename);
    if(!in)
        throw(string("Can't open ")+filename);
    string strLine;
    while(getline(in,strLine))
    {
        vector<string> parsed = Tokenize(strLine);
        // C*12:03:01G	C*12:03:01:01	C*12:03:01:02	C*12:03:06	C*12:23
        // C*14:02:01G	C*14:02:01	C*14:02:07	C*14:23	C*14:31
        // C*15:02:01G	C*15:02:01	C*15:02:07	C*15:13	C*15:47
        if(parsed.size() >= 2)
        {
            string new_name = parsed[0];
            vector<string> alleles;
            for(int i = 1; i < parsed.size(); i++)
            {
                if(parsed[i] != "")
                    alleles.push_back(parsed[i]);
            }
            m_vctAmbiguities.push_back(CGAllele(new_name,alleles));
        }
    }
    in.close();

    // Allele,pair<m_vctAmbiguities[idx], m_vctAmbiguities[idx].alleles[idx]>
    for(size_t g = 0; g < m_vctAmbiguities.size(); g++)
        for(size_t i = 0; i < m_vctAmbiguities[g].alleles.size();i++)
        {
            m_amb_pointer[m_vctAmbiguities[g].alleles[i]] = pair<size_t,size_t>(g,i);
        }
}

vector<string> CImgtAmbiguities::Tokenize(const string& val)
{
    return Tokenize(val,"\t");
}

vector<string> CImgtAmbiguities::Tokenize(const string& str, const string& delimiters = "\t")
{
    vector<string> tokens;
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        if(str.length() > pos && delimiters.find(str[pos]) != string::npos )
            lastPos=pos+1;
        else
            lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
    return tokens;
}

bool CImgtAmbiguities::isAmbiguityOfBetterGAllele(const string& allele,vector<string> alternatives)
{
    map<string,pair<size_t,size_t> >::iterator iter = m_amb_pointer.find(allele);

    if(iter == m_amb_pointer.end())
        return false;

    size_t gIdx         = iter->second.first;
    size_t order_Idx    = iter->second.second;

    for(size_t i = 0; i < alternatives.size(); i++)
    {
        map<string,pair<size_t,size_t> >::iterator iter = m_amb_pointer.find(alternatives[i]);
        //cout << iter->second.first << " " << iter->second.second << endl;
        if(iter != m_amb_pointer.end() && iter->second.first == gIdx && iter->second.second < order_Idx)
            return true;
    }
    return false;
}


