/* 
 * File:   CTwoBit.h
 * Author: mwittig
 *
 * Created on February 18, 2011, 10:52 AM
 */

#ifndef _CTWOBIT_H
#define	_CTWOBIT_H

using namespace std;

class CTwoBit {

    typedef struct{
        unsigned int signature;
        unsigned int version;
        unsigned int seqCount;
        unsigned int reserved;
    }header;

    typedef struct{
        unsigned int            offset;
        unsigned int            dnaSize;
        unsigned int            nBlockCount;
        vector<unsigned int>    nBlockStarts;
        vector<unsigned int>    nBlockSizes;
        unsigned int            maskBlockCount;
        vector<unsigned int>    maskBlockStarts;
        vector<unsigned int>    maskBlockSizes;
        unsigned int            reserved;
        char*               packedDNA;
    }chrom_data;


public:
    CTwoBit();
    CTwoBit(const char*);
    CTwoBit(const CTwoBit& orig);
    virtual ~CTwoBit();
    
    string getSequence(const char* chrom,unsigned int start, unsigned int end, unsigned int flanks);
    string getSequence(const char* chrom,unsigned int start, unsigned int end);
    string getSequence(const char* chrom);
    string getFaSequence(const char* chrom,unsigned int start, unsigned int end,unsigned int lineBreaks = 50);
    string getFaSequence(const char* chrom,unsigned int start, unsigned int end,const char* faHeader ,unsigned int lineBreaks = 50);

    vector<pair<string,unsigned int> >  getChromSizes();

private:
    bool        m_bOpen;
    ifstream*   m_pFile;
    header      m_stcHeader;

    map<string,chrom_data> m_data;

    bool storeOffsets();
    bool storeData();
    char getAllele(unsigned int,unsigned int);
    void nBlocks(ostringstream& strSequence,map<string,chrom_data>::const_iterator iter,unsigned int start,unsigned int end);
    
    bool loadTwoBit(const char*);
    void closeTwoBit();

    void initialize(){
        m_bOpen = false;
        m_pFile = NULL;
    }
    
    chrom_data getChromDataStruc(){
        chrom_data cdReturn;
        cdReturn.dnaSize = 0;
        cdReturn.maskBlockCount = 0;
        cdReturn.nBlockCount = 0;
        cdReturn.offset = 0;
        cdReturn.packedDNA = NULL;
        cdReturn.reserved = 0;
        return cdReturn;
    }

};

#endif	/* _CTWOBIT_H */

