/* 
 * File:   main.cpp
 * Author: mwittig
 *
 * Created on July 19, 2012, 2:39 PM
 */

#include <cctype>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <exception>
#include <vector>
#include <list>
#include <map>
#include <utility>
#include <string>
#include <iomanip>
#include <iostream>
#include <sstream>
#include <bitset>
#include <fstream>
#include <algorithm>
#include <set>

#include "CMultiFasta.h"
#include "CMWException.h"

#include "CBlat.h"
#include <boost/regex.hpp>
#include <boost/regex.h>
#include <iosfwd>
#include <streambuf>
#include <math.h>

#include "CImgtDb.h"


#include "CHLAAlignment.h"
#include "CImgtAmbiguities.h"


// #define SCRATCH "/home/mwittig/tmpfs/VBox.tmpfs"
// #define HLA_IMGT_DB "/home/mwittig/data/HLA/WGS/brute/B_nuc.fasta"

using namespace std;

// Wieviel % der exon coverage muss am der Exon Grenze sein
#define FACTOR_EXON_BOUNDARY_COVERAGE 0.1
#define BOOST_ALL_NO_LIB

class  stc_mapping_results{
    public:
        stc_mapping_results()
        {allele_seq_size = alignable_seq_size = covered_bases_count = aligned_read_count = paired_read_count = exon_boundary_penalties = coverage_drop_penalties = 0; area_under_curve_error = 0.0f; allele = "";}
        stc_mapping_results(const stc_mapping_results& orig){
            allele                  = orig.allele;
            allele_seq_size         = orig.allele_seq_size;
            alignable_seq_size      = orig.alignable_seq_size;
            covered_bases_count     = orig.covered_bases_count;
            aligned_read_count      = orig.aligned_read_count;
            paired_read_count       = orig.aligned_read_count;
            exon_boundary_penalties = orig.exon_boundary_penalties;
            coverage_drop_penalties = orig.coverage_drop_penalties;
            paired_reads_detailed   = orig.paired_reads_detailed;
            single_reads_detailed   = orig.single_reads_detailed;
            area_under_curve_error  = orig.area_under_curve_error;
        }
        ~stc_mapping_results(){ }

        string  allele;
        unsigned long   allele_seq_size;
        unsigned long   alignable_seq_size;
        unsigned long   covered_bases_count;
        unsigned long   aligned_read_count;
        unsigned long   paired_read_count;
        unsigned long   exon_boundary_penalties;
        unsigned long   coverage_drop_penalties;
        double          area_under_curve_error;
        map<string, pair<CBlat::stcAlignment,CBlat::stcAlignment> >     paired_reads_detailed;
        map<string, CBlat::stcAlignment>                                single_reads_detailed;
};


class  stc_reads_per_allele{
    public:
        stc_reads_per_allele(){A_reads = A_pairs = B_reads = B_pairs =  0;}
        stc_reads_per_allele(const stc_reads_per_allele& orig){
            A_reads = orig.A_reads;
            A_pairs = orig.A_pairs;
            B_reads = orig.B_reads;
            B_pairs = orig.B_pairs;
        }
        ~stc_reads_per_allele(){}
        unsigned long A_reads;
        unsigned long A_pairs;
        unsigned long B_reads;
        unsigned long B_pairs;
};

class stc_paired_alleles {
    public:
        stc_paired_alleles() : Fscore(0.0f), read_equality(0.0f), read_per_base(0.0f), pairs_per_read(0.0f), redundant_read_per_base(0.0f), area_under_curve_error(0){}
        stc_paired_alleles(const stc_paired_alleles& orig) : Fscore(0.0f){
            A       = orig.A;
            B       = orig.B;
            reads   = orig.reads;
            Fscore  = orig.Fscore;
            read_equality = orig.read_equality;
            read_per_base = orig.read_per_base;
            pairs_per_read= orig.pairs_per_read;
            redundant_read_per_base=orig.redundant_read_per_base;
            area_under_curve_error = orig.area_under_curve_error;
        }
        ~stc_paired_alleles(){}

        stc_mapping_results     A; // hier sind die read und pair counts redundant
        stc_mapping_results     B; // hier sind die read und pair counts redundant
        stc_reads_per_allele    reads; // hier sind alle Redundanzen beseitigt
        double                  Fscore;
        double                  read_equality;
        double                  read_per_base;
        double                  pairs_per_read;
        double                  redundant_read_per_base;
        double                  area_under_curve_error;
};

// Border mapping bases are required to allow softclipping at sequence start and sequence end

int cmd(const string& commando, bool write_output=true);
string itoa(int value, int base);
string ftoa(double val);
vector<string> Tokenize(const string&, const string& delimiters);
vector<string> Tokenize(const string&);

string erase_characters (string rep, string in );
string string_replace( string src, string const& target, string const& repl);

vector<pair<unsigned int, unsigned int> > getExonBlocks(const string&);

multimap<long,long> getSortedReads(map<string,CBlat::stcAlignment>& the_map);
multimap<long,long> getSortedReads(multimap<string,CBlat::stcAlignment>& the_map);

stc_mapping_results getMappingResults(const string& allele, const string& sequence, const string& scratch_reference, const string& scratch_psl,
                                      const string& scratch_plot_reads, const string& scratch_plot_coverage, const string& scratch_plot_cmd, const string& scratch_plot_boundaries,
                                      const string& arg_reads, const string& arg_result, const string& arg_scratch, const string& arg_plots, const long& arg_read_length,
                                      bool return_read_maps = false);

stc_paired_alleles getAlleleSpecificReads(const stc_mapping_results& A,const stc_mapping_results& B);
stc_mapping_results getReducedMappingResults(const stc_mapping_results& in);

// sorting Funktion für vector mit mapping Ergebnissen
// Sortierregeln für cDNA
bool sortCDNAmappingResults(const stc_mapping_results&,const stc_mapping_results&);
// gibt nur die mappings zurück, die als relevante Hits gelten
vector<stc_mapping_results> filterCdnaMappingResults(const vector<stc_mapping_results>&);
vector<stc_mapping_results> filterGdnaMappingResults(const vector<stc_mapping_results>&);

void calculateMWScore(vector<stc_paired_alleles>&);

bool sortAllelePairingResults(const stc_paired_alleles&,const stc_paired_alleles&);
vector<string> removeAmbiguityCandidates(vector<stc_mapping_results>&,CImgtAmbiguities&);
void removeShortCDNAbias(vector<stc_mapping_results>&,map<string,string>&, bool prefer_longer_annotation);

vector<string> getAlternativeAlleles(const vector<stc_paired_alleles>&);
string getAllelesAstring(const vector<string>&, unsigned int resolution = 3);
void removeAllelesWhichAreEquivalent(const string& allel,vector<string>& alleles, unsigned int resolution = 3);

// Vergleiche Namensgebung. Ist A ursprünglicher als b? Mehr Felder ist Ursprünglicher. Bei gleichen Feldern ist von vorn nach hinten die kleinere Nummer ursprünglicher
bool isAncientAnnotation(const string& a, const string& b);
bool isTheSameAtTwoDigit(const string& a, const string& b);
string getAlleleName(const string& allele,unsigned int resolution = 3);

void bringUpTheMostAncientAllleleCombination(vector<stc_paired_alleles>&,double f_score_within = 0.8);
// Sind die reads pro base zu ungleich verteilt, dann ist es sehr wahrscheinlich das Homozygote allele. Grenzwert ist das 3 fache
void bringUpTheMostAncientHomozygousAllele(vector<stc_paired_alleles>& allele_gt,double inequality_factor = 3);

void eraseRedundantSequences(map<string,string>&);


size_t countUniqueAlleleSpecificReads(map<string, CBlat::stcAlignment> reads, vector<unsigned int> exon_list);
size_t countUniqueAlleleSpecificReads(map<string, CBlat::stcAlignment> reads);

int main(int argc, char** argv)
{
    try
    {
        /*
          //  COMMAND LINE PARSER
         *
         *
         */
        
        string arg_cDNA     = "";
        string arg_gDNA     = "";
        string arg_scratch  = "";
        string arg_result   = "";
        string arg_plots    = "";
        string arg_reads    = "";
        string arg_bias     = "";
        string arg_ambiguity= "";
        long   arg_read_length= 100;

        if(argc != 19)
        { // Check the value of argc. If not enough parameters have been passed, inform user and exit.
            std::cout << "Usage is:" << endl << " --cDNA <infile> --gDNA <infile> --scratch <path> --out <outfile> -outPic <path> --bias-alleles <infile> --amb-table <infile> --reads <infile>" << endl;
            exit(EXIT_FAILURE);
        }
        else
        {
            // if we got enough parameters...
            char* myFile, myPath, myOutPath;
            std::cout << "  starting " << argv[0] << endl << "  arguments:" << endl;
            for (int i = 1; i < argc-1; i+=2)
            {
                if(string(argv[i]).compare("--cDNA")==0)
                    arg_cDNA = argv[i+1];
                else if(string(argv[i]).compare("--gDNA")==0)
                    arg_gDNA = argv[i+1];
                else if(string(argv[i]).compare("--scratch")==0)
                    arg_scratch = argv[i+1];
                else if(string(argv[i]).compare("--out")==0)
                    arg_result= argv[i+1];
                else if(string(argv[i]).compare("--outPic")==0)
                    arg_plots = argv[i+1];
                else if(string(argv[i]).compare("--reads")==0)
                    arg_reads = argv[i+1];
                else if(string(argv[i]).compare("--bias-alleles")==0)
                    arg_bias = argv[i+1];
                else if(string(argv[i]).compare("--amb-table")==0)
                    arg_ambiguity = argv[i+1];
                else if(string(argv[i]).compare("--read-length")==0)
                    arg_read_length = atol(argv[i+1]);
                else
                    throw(string("Unexpected argument ")+argv[i]+", check spelling !!!");
                cout << setw(11) << right << argv[i] << ' ' << argv[i+1] << endl;

            }
         }

        /*  BLOCK 1 , Vorbereitung
         * #################################################
         *  Lese als erstes die gDNA und cDNA Sequenzen
         *  Lösche alle cDNA Sequenzen von denen gDNA Sequenz verfügbar
         *  Lösche alle Sonderzeihen in den gDNA Sequenzen
         *  Losche * und . in den cDNA Sequenzen und alle leading & trailing |
         *  Packe beides zusammen zu einer Sequenzkollektion
         */
        // Arbeite folgende loci ab
        //string gen_names[] = {"A","B","C","DMA","DMB","DOA","DOB","DPA","DPB","DQA","DQB","DRA","DRB3","DRB4","DRB","E","F","G","H","J","K","L","MICA","MICB","P","TAP1","TAP2","V"};
        //string nuc_names[] = {"A","B","C","DMA","DMB","DOA","DOB","DPA","DPB","DQA","DQB1","DQB","DRA","DRB1","DRB3","DRB","E","F","G","H","J","K","L","MICA","MICB","TAP1","TAP2","V"};
        
        // Initialize true random
        srand(time(0));

        // Lese als erstes die gDNA und cDNA Sequenzen
        cout  << endl << "Reading gDNA and cDNA sequences, bias alleles and ambiguities table ... ";
        //CImgtDb idbG(arg_gDNA.c_str());
        //CImgtDb idbC(arg_cDNA.c_str());
        CImgtDb idbC(arg_cDNA.c_str(),arg_bias.c_str());
        CImgtAmbiguities imgt_ambiguities(arg_ambiguity.c_str(),"");

        // map<string,string> mG = idbG.getAll();   // map mit gDNA Sequenzen
        map<string,string> mC = idbC.getAll();   // map mit cDNA Sequenzen

        cout << "done. " << endl << ", cDNA: " << mC.size() << endl;
        // cout << "done." << endl << "genomic: " << mG.size() << ", cDNA: " << mC.size() << endl;
        
        // Lösche alle cDNA Sequenzen von denen gDNA Sequenz verfügbar
        // Lösche alle Sonderzeihen in den gDNA Sequenzen
        /*
        for(map<string,string>::iterator iter = mG.begin(); iter != mG.end(); iter++)
        {
             //mC.erase(iter->first);
             iter->second = erase_characters("*.|", iter->second);
        }
        */
        // Losche * und . in den cDNA Sequenzen und alle leading & trailing |
        for(map<string,string>::iterator iter = mC.begin(); iter != mC.end(); iter++)
        {
            iter->second = erase_characters("*.", iter->second);
            iter->second = string_replace(iter->second,"||","|");
            while(iter->second[0] == '|')
                iter->second.erase(iter->second.begin());
            while(iter->second[iter->second.size()-1] == '|')
                iter->second = iter->second.substr(0,iter->second.size()-1);
        }

        eraseRedundantSequences(mC);
        // eraseRedundantSequences(mG);

        /* Ende Block 1
         * #########################################
         */
        string scratch_reference        = arg_scratch + "/temp_ref.fasta";
        string scratch_psl              = arg_scratch + "/blat_out.psl";
        string scratch_plot_reads       = arg_scratch + "/reads_aligned.txt";
        string scratch_plot_coverage    = arg_scratch + "/coverage.txt";
        string scratch_plot_boundaries  = arg_scratch + "/ablines.txt";
        string scratch_plot_cmd         = arg_scratch + "/plot.R";
        
        ofstream final_output(arg_result.c_str());
        if(!final_output)
            throw(string("Can't open ")+arg_result);
        // final_output << "locus\tseq_length\talignable_sequence\tcovered_bases\taligned_reads\taligned_pairs\texon_boundary_penalties\tcoverage_drop_penalty" << endl;
        final_output << "A.allele\tA.allele_seq_size\tA.alignable_seq_size\tA.covered_bases_count\treads.A_reads\treads.A_pairs\tA.exon_boundary_penalties\tA.coverage_drop_penalties\tB.allele\tB.allele_seq_size\tB.alignable_seq_size\tB.covered_bases_count\treads.B_reads\treads.B_pairs\tB.exon_boundary_penalties\tB.coverage_drop_penalties\tread_per_base\tpairs_per_read\tread_equality\tarea_under_curve_error\tFscore" << endl;
        final_output.close();

        /* Block 2, Alignment gegen cDNA
         * 
         */
        vector<stc_mapping_results> cDNA_mapping;
        bool debug = false;
        for(map<string,string>::iterator itercDNA = mC.begin(); itercDNA != mC.end(); itercDNA++)
        {
            /*
            // ============ =========== DEBUG ============ ===========
            // debug a specific sequence
            if(itercDNA->first.compare("B*40:01:01") == 0)
                debug = true;
            if(!debug)
                continue;
            // */
            //*
            // ============ =========== DEBUG ============ ===========
            //if(itercDNA->first.compare("B*35:42:01") != 0 &&  itercDNA->first.compare("C*14:02:01") != 0 &&  itercDNA->first.compare("C*04:09N") != 0)
            //    continue;
            /*
            if( itercDNA->first.compare("A*32:01:01") != 0 &&  itercDNA->first.compare("A*02:01:01:01") != 0 &&  itercDNA->first.compare("A*03:01:01:01") != 0 &&  itercDNA->first.compare("A*03:12") != 0 &&
                itercDNA->first.compare("A*02:135") != 0 &&  itercDNA->first.compare("A*32:12") != 0 &&  itercDNA->first.compare("A*02:121") != 0 &&  itercDNA->first.compare("A*26:15") != 0)
                continue;
            // */

            stc_mapping_results the_result = getMappingResults(itercDNA->first, itercDNA->second, scratch_reference, scratch_psl,
                                      scratch_plot_reads, scratch_plot_coverage, scratch_plot_cmd, scratch_plot_boundaries,
                                      arg_reads, arg_result, arg_scratch, arg_plots, arg_read_length);

            cDNA_mapping.push_back(the_result);
            // cout << "locus\tseq_length\talignable_sequence\tcovered_bases\taligned_reads\taligned_pairs\texon_boundary_penalties\tcoverage_drop_penalty" << endl;
            //*
            final_output.open(arg_result.c_str(),ios::app);
            if(!final_output)
                throw(string("Can't open ")+arg_result);
            final_output<< the_result.allele << '\t'
                        << the_result.allele_seq_size << '\t'
                        << the_result.alignable_seq_size << '\t'
                        << the_result.covered_bases_count << '\t'
                        << the_result.aligned_read_count << '\t'
                        << the_result.paired_read_count << '\t'
                        << the_result.exon_boundary_penalties << '\t'
                        << the_result.coverage_drop_penalties << endl;
            final_output.close();
            // */
        }

        // clean and order best mappings
        vector<stc_mapping_results> best_cDNA_hits = filterCdnaMappingResults(cDNA_mapping);
        vector<string> g_alleles_removed = removeAmbiguityCandidates(best_cDNA_hits,imgt_ambiguities);
        sort(best_cDNA_hits.begin(),best_cDNA_hits.end(),sortCDNAmappingResults);
        removeShortCDNAbias(best_cDNA_hits,mC,true);

        // remap und speichere die read ids für die paired reads und die unpaired reads
        //for(vector<stc_mapping_results>::iterator iter = best_cDNA_hits.begin(); iter != best_cDNA_hits.end(); iter++)
        for(int i = 0; i < best_cDNA_hits.size(); i++)
        {
            string seq = mC[best_cDNA_hits[i].allele];
            // Die nächsten beiden Zeilen nutzen gDNA statt cDNA, wenn verfügbar
            // if(mG.find(best_cDNA_hits[i].allele) != mG.end()) // only cDNA Sequenz verfügbar
            //    seq = mG[best_cDNA_hits[i].allele];
            best_cDNA_hits[i] = getMappingResults(best_cDNA_hits[i].allele, seq, scratch_reference, scratch_psl,
                                      scratch_plot_reads, scratch_plot_coverage, scratch_plot_cmd, scratch_plot_boundaries,
                                      arg_reads, arg_result, arg_scratch, arg_plots, arg_read_length,true);
        }
        // gDNA filtering, discard coverage drops
        // best_cDNA_hits = filterGdnaMappingResults(best_cDNA_hits);

        vector<stc_paired_alleles>  allele_gt;
        if(best_cDNA_hits.size() > 0)
        {
            for(vector<stc_mapping_results>::iterator iterA = best_cDNA_hits.begin(); iterA != best_cDNA_hits.end(); iterA++)
                for(vector<stc_mapping_results>::iterator iterB = iterA; iterB != best_cDNA_hits.end(); iterB++)
                {
                    // cout << iterA->allele << '\t' << iterB->allele << endl;
                    // ACHTUNG true ist hier ein Versuch abgeschnittene reads zu Ignorieren
                    stc_paired_alleles pa = getAlleleSpecificReads(*iterA,*iterB);
                    allele_gt.push_back(pa);
                }
        }
        //calculateF1Score(allele_gt);
        if(allele_gt.size() != 0)
        {
            calculateMWScore(allele_gt);
            sort(allele_gt.begin(),allele_gt.end(),sortAllelePairingResults);
            // Achtung, war sonst immer auf 0.75
            bringUpTheMostAncientAllleleCombination(allele_gt,0.5);
            bringUpTheMostAncientHomozygousAllele(allele_gt,3);
            

            //* paired allele output
            //*
            final_output.open(arg_result.c_str(),ios::app);
                if(!final_output)
                    throw(string("Can't open ")+arg_result);

            for(int i = 0; i < allele_gt.size(); i++)
                final_output << allele_gt[i].A.allele << '\t'
                             << allele_gt[i].A.allele_seq_size << '\t'
                             << allele_gt[i].A.alignable_seq_size << '\t'
                             << allele_gt[i].A.covered_bases_count << '\t'
                             << allele_gt[i].reads.A_reads << '\t'
                             << allele_gt[i].reads.A_pairs << '\t'
                             << allele_gt[i].A.exon_boundary_penalties << '\t'
                             << allele_gt[i].A.coverage_drop_penalties << '\t'
                             << allele_gt[i].B.allele << '\t'
                             << allele_gt[i].B.allele_seq_size << '\t'
                             << allele_gt[i].B.alignable_seq_size << '\t'
                             << allele_gt[i].B.covered_bases_count << '\t'
                             << allele_gt[i].reads.B_reads << '\t'
                             << allele_gt[i].reads.B_pairs << '\t'
                             << allele_gt[i].B.exon_boundary_penalties << '\t'
                             << allele_gt[i].B.coverage_drop_penalties << '\t'
                             << allele_gt[i].read_per_base << '\t'
                             << allele_gt[i].pairs_per_read << '\t'
                             << allele_gt[i].read_equality << '\t'
                             << allele_gt[i].area_under_curve_error << '\t'
                             << allele_gt[i].Fscore << endl;

            // Stelle die Liste der Alternativen Möglichkeiten zusammen und lösche alle Redundanzen in dieser
            vector<string> alt_alleles = getAlternativeAlleles(allele_gt);
            alt_alleles.insert(alt_alleles.end(),g_alleles_removed.begin(),g_alleles_removed.end());
            removeAllelesWhichAreEquivalent(allele_gt[0].A.allele,alt_alleles,3);
            removeAllelesWhichAreEquivalent(allele_gt[0].B.allele,alt_alleles,3);
            string alternative_alleles_entry = getAllelesAstring(alt_alleles, 3);
            if(alternative_alleles_entry == "")
                alternative_alleles_entry = "-";
            
            final_output << "#[Result]" << endl;
            final_output << getAlleleName(allele_gt[0].A.allele,3) << '\t'
                         << getAlleleName(allele_gt[0].B.allele,3) << '\t'
                         << alternative_alleles_entry << '\t'
                         << allele_gt[0].Fscore
                         << endl;
            final_output.close();
        //*/
        }
        return (EXIT_SUCCESS);
    }
    catch(CMWException ex)
    {
        cerr << "Expected error: " << ex.what() << endl;
    }
    catch(exception ex)
    {
        cerr << "Expected error: " << ex.what() << endl;
    }
    catch(string ex)
    {
        cerr << "Expected error: " << ex << endl;
    }
    catch(...)
    {
        cerr << "unexpected error. closing ..." << endl;
    }
}

int cmd(const string& commando, bool write_output)
{
    FILE *in;
    char buff[512];

    if(!(in = popen(commando.c_str(), "r")))
    {
        return 1;
    }

    while(fgets(buff, sizeof(buff), in)!=NULL)
        if(write_output)
            cout << buff;
    pclose(in);
    if(write_output)
        cout << endl;
    return 0;
}

string itoa(int value, int base)
{

    string buf;

    // check that the base if valid
    if (base < 2 || base > 16) return buf;

    enum { kMaxDigits = 35 };
    buf.reserve( kMaxDigits ); // Pre-allocate enough space.

    int quotient = value;

    // Translating number to string with base:
    do {
            buf += "0123456789abcdef"[ std::abs( quotient % base ) ];
            quotient /= base;
    } while ( quotient );

    // Append the negative sign
    if ( value < 0) buf += '-';

    std::reverse( buf.begin(), buf.end() );
    return buf;
}

vector<string> Tokenize(const string& val)
{
    return Tokenize(val,"\t");
}

vector<string> Tokenize(const string& str, const string& delimiters = "\t")
{
    vector<string> tokens;
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        if(str.length() > pos && delimiters.find(str[pos]) != string::npos )
            lastPos=pos+1;
        else
            lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
    return tokens;
}

string string_replace( string src, string const& target, string const& repl)
{
    // handle error situations/trivial cases

    if (target.length() == 0) {
        // searching for a match to the empty string will result in
        //  an infinite loop
        //  it might make sense to throw an exception for this case
        return src;
    }

    if (src.length() == 0) {
        return src;  // nothing to match against
    }

    size_t idx = 0;

    for (;;) {
        idx = src.find( target, idx);
        if (idx == string::npos)  break;

        src.replace( idx, target.length(), repl);
        idx += repl.length();
    }

    return src;
}

string ftoa(double val)
{
    char temp[100];
    sprintf(temp,"%.12f",val);
    cout << temp << endl;
    return string(temp);
}

string erase_characters (string rep, string in)
{
  int pos;
  while (true)
  {
    pos = in.find_first_of(rep);
    if (pos == -1)
      break;
    else
      in.erase(pos, 1);
  }
  return in;
}


multimap<long,long> getSortedReads(map<string,CBlat::stcAlignment>& the_map)
{
    multimap<long,long> mapReturn;
    for(map<string,CBlat::stcAlignment>::const_iterator iter = the_map.begin(); iter != the_map.end(); iter++)
        mapReturn.insert(pair<long,long>(iter->second.RefStart,iter->second.RefStart+iter->second.AlignLength-1));
    return mapReturn;
}

multimap<long,long> getSortedReads(multimap<string,CBlat::stcAlignment>& the_map)
{
    multimap<long,long> mapReturn;
    for(multimap<string,CBlat::stcAlignment>::const_iterator iter = the_map.begin(); iter != the_map.end(); iter++)
        mapReturn.insert(pair<long,long>(iter->second.RefStart,iter->second.RefStart+iter->second.AlignLength-1));
    return mapReturn;
}

stc_mapping_results getMappingResults(const string& allele, const string& sequence, const string& scratch_reference, const string& scratch_psl,
                                      const string& scratch_plot_reads, const string& scratch_plot_coverage, const string& scratch_plot_cmd, const string& scratch_plot_boundaries,
                                      const string& arg_reads, const string& arg_result, const string& arg_scratch, const string& arg_plots, const long& arg_read_length,
                                      bool return_read_maps)
{
    CHLAAlignment alignment(sequence,arg_read_length);
    // cout << alignment.getExonReport();

    // store reference for blat processing
    alignment.getExonsAsFasta(scratch_reference);

    // run blat
    CBlat blat;
    blat.run(scratch_reference,arg_reads,scratch_psl,arg_read_length);

    //cout << setw(30) << left << itercDNA->first;
    //cout << blat.m_aligned_reads.size() << " reads, removing truncated reads ... ";
    blat.removeTruncatedReads(alignment.m_exon_sequences,arg_read_length); // Hiernach werden auch die paired reads gezählt !!!!
    //cout << blat.m_aligned_reads.size() << " reads left, removing duplicates ...";
    blat.removeRedundantAlignments();
    //cout << blat.m_aligned_reads.size() << " reads left." << endl;
    alignment.transferReadsFromBlat(blat.m_aligned_reads);
    // Errechnet central reads usw.
    alignment.calculateCoverageMetrics();

    string scratch_plot_coverage_data = string(scratch_plot_coverage)+".data.txt";
    string scratch_plot_coverage_model = string(scratch_plot_coverage)+".model.txt";
    string scratch_plot_central_reads = string(scratch_plot_coverage)+".central.txt";
    string scratch_plot_left_right_reads = string(scratch_plot_coverage)+".leftright.txt";
    string scratch_plot_score_failed = string(scratch_plot_coverage)+".score_failed.txt";

    alignment.getSortedReads(scratch_plot_reads, true);
    alignment.getCoverage(scratch_plot_coverage_data);
    alignment.getIdealCoverage(scratch_plot_coverage_model);
    alignment.getExonBoundaries(scratch_plot_boundaries);
    alignment.getCentralReadsScore(scratch_plot_central_reads);
    alignment.getLeftRightScore(scratch_plot_left_right_reads);
    alignment.getScoreFailed(scratch_plot_score_failed);

    // cout << "locus\tseq_length\talignable_sequence\tcovered_bases\taligned_reads\taligned_pairs\texon_boundary_penalties\tcoverage_drop_penalty" << endl;
    
    stc_mapping_results stcReturn;

    stcReturn.allele                    = allele;
    stcReturn.allele_seq_size           = alignment.getSequenceSize();
    stcReturn.alignable_seq_size        = alignment.getAlignableSequenceSize(30);
    stcReturn.covered_bases_count       = alignment.getNumberOfCoveredBases();
    stcReturn.aligned_read_count        = alignment.m_aligned_reads.size();
    stcReturn.paired_read_count         = blat.pairedReadCount();
    stcReturn.exon_boundary_penalties   = alignment.getExonBoundaryPenalties(alignment.Mean,FACTOR_EXON_BOUNDARY_COVERAGE);
    //stcReturn.coverage_drop_penalties   = alignment.findCoverageDrops().size();
    stcReturn.coverage_drop_penalties   = alignment.findCoverageDropsExtended().size();
    stcReturn.area_under_curve_error    = alignment.getAreaUnderCurveError();
    if(return_read_maps)
    {
        stcReturn.paired_reads_detailed = blat.m_paired_reads_detailed;
        stcReturn.single_reads_detailed = blat.m_unpaired_reads_detailed;

    }
    
    // Automatischer R plot für Testszwecke
    ofstream out(scratch_plot_cmd.c_str());
    out << "setwd(\"" << arg_scratch << "/\")" << endl;
    out << "a=read.table(\"" << scratch_plot_coverage_data << "\",header=F)" << endl;
    out << "cr=read.table(\"" << scratch_plot_central_reads << "\",header=F)" << endl;
    out << "lr=read.table(\"" << scratch_plot_left_right_reads << "\",header=F)" << endl;
    out << "sf=NULL" << endl;
    out << "try({sf=read.table(\"" << scratch_plot_score_failed << "\",header=F)},silent=TRUE)" << endl;
    out << "ab=read.table(\"" << scratch_plot_coverage_model << "\",header=F)" << endl;
    out << "b=read.table(\"" << scratch_plot_reads << "\",header=F,colClasses=c(\"integer\"))" << endl;
    out << "c=read.table(\"" << scratch_plot_boundaries << "\",header=F,colClasses=c(\"integer\"))" << endl;
    out << "png(filename = \"" << arg_plots << "/" << string_replace(allele,"*","-") << ".png\",width = 1024, height = 768, pointsize = 20)" << endl;
    out << "nf = layout(matrix(c(1,2),2,1,byrow=TRUE), heights=c(1,1), TRUE)" << endl;
    out << "par(mar=c(-.1, 4, 3, .2) + 0.1)" << endl;
    out << "max_x = max(c(ab[,1],b[,3]))" << endl;
    out << "d = which(a[,2] == 0)" << endl;
    out << "if(length(d) > 0){a = a[-d,]}" << endl;
    out << "d = which(ab[,2] == 0)" << endl;
    out << "if(length(d) > 0){ab = ab[-d,]}" << endl;
    out << "plot(NULL, type=\"h\",xlab=NULL,ylab=\"coverage\",xlim=c(0,max_x),ylim=c(0,100),main=\"" << allele << "\",xaxt=\"n\",yaxt=\"n\",bty =\"n\")" << endl;
    out << "bb <- par(\"usr\")" << endl;
    out << "abline(v=bb[1:2],h=bb[3:4],lwd=1,col=\"#000000\")" << endl;
    out << "axis(2,at=seq(0,100,20),labels=seq(0,100,20),las=2 )" << endl;
    out << "abline(v=c[,1],col=\"#9999ff\")" << endl;
    out << "rect(xleft=ab[,1], xright=ab[,1], ybottom=0,  ytop=ab[,2],col=\"#ff9999\", border=\"#ff9999\"  )" << endl;
    out << "if( NROW(a) > 0 ){rect(xleft=a[,1], xright=a[,1], ybottom=0,  ytop=a[,2] )}" << endl;
    out << "par(mar=c(4, 4, -.1, .2) + 0.1)" << endl;
    //out << "cr[-which(cr[,2]<0),2]=0 "<< endl;
    out << "cr[,2]=cr[,2]*10" << endl;
    //out << "lines(cr,col=\"#A901DB\")" << endl;
    out << "plot(NULL,xlim=c(0,max_x),ylim=c(0,100),ylab=\"reads\",xlab=\"coordinate\",yaxt=\"n\",bty =\"n\")" << endl;
    out << "bb <- par(\"usr\")" << endl;
    out << "abline(h=0.2*10,col=\"red\")" << endl;
    out << "abline(v=bb[1:2],h=bb[3],lwd=1,col=\"#000000\")" << endl;
    out << "axis(2,at=seq(0,100,20),labels=seq(0,100,20),las=2)" << endl;
    out << "abline(v=c[,1],col=\"#9999ff\")" << endl;
    out << "if( NROW(b) > 0 ){segments(b[,1],b[,2],b[,3],b[,4],lwd =1,col=\"black\")}" << endl;
    //out << "lr[-which(lr[,2]<0),2]=0 "<< endl;
    out << "lr[,2]=lr[,2]*10" << endl;
    out << "lines(cr,col=\"#A901DB\")" << endl;
    out << "if(NROW(sf) > 0){axis(1,at=sf[,1],labels=NA,col=\"red\")}" << endl;
    out << "" << endl;
    out << "dev.off()" << endl;
    out.close();

    cmd(string("R --vanilla < ")+scratch_plot_cmd, false);

    return stcReturn;
}

bool sortCDNAmappingResults(const stc_mapping_results& a,const stc_mapping_results& b)
{
    if(a.covered_bases_count/a.alignable_seq_size > b.covered_bases_count/b.alignable_seq_size)
        return true;
    else if(a.covered_bases_count/a.alignable_seq_size < b.covered_bases_count/b.alignable_seq_size)
        return false;
    if(a.exon_boundary_penalties < b.exon_boundary_penalties)
        return true;
    else if(a.exon_boundary_penalties > b.exon_boundary_penalties)
        return false;
    if(a.coverage_drop_penalties < b.coverage_drop_penalties)
        return true;
    else if(a.coverage_drop_penalties > b.coverage_drop_penalties)
        return false;
    if(a.paired_read_count > b.paired_read_count)
        return true;
    else if(a.paired_read_count < b.paired_read_count)
        return false;
    if(a.aligned_read_count > b.aligned_read_count)
        return true;
    else if(a.aligned_read_count < b.aligned_read_count)
        return false;
    if(a.alignable_seq_size > b.alignable_seq_size)
        return true;
    else if(a.alignable_seq_size < b.alignable_seq_size)
        return false;
    return true;

}

vector<stc_mapping_results> filterCdnaMappingResults(const vector<stc_mapping_results>& in_vector)
{
    vector<stc_mapping_results> out_vector;

    for(vector<stc_mapping_results>::const_iterator iter = in_vector.begin();iter != in_vector.end(); iter++)
    {
        //if(iter->exon_boundary_penalties == 0 && iter->coverage_drop_penalties == 0 && iter->alignable_seq_size == iter->covered_bases_count)
        if(iter->coverage_drop_penalties == 0 && iter->alignable_seq_size == iter->covered_bases_count)
            out_vector.push_back(*iter);
    }
    return out_vector;
}

vector<stc_mapping_results> filterGdnaMappingResults(const vector<stc_mapping_results>& in_vector)
{
    vector<stc_mapping_results> out_vector;

    for(vector<stc_mapping_results>::const_iterator iter = in_vector.begin();iter != in_vector.end(); iter++)
    {
        if(iter->coverage_drop_penalties == 0)
            out_vector.push_back(*iter);
    }
    return out_vector;
}

/* Old Style
stc_paired_alleles getAlleleSpecificReads(const stc_mapping_results& A,const stc_mapping_results& B)
{
    stc_reads_per_allele stcPA;
    
    for(map<string, CBlat::stcAlignment>::const_iterator iter = A.single_reads_detailed.begin(); iter != A.single_reads_detailed.end(); iter++)
        if(B.single_reads_detailed.find(iter->first) == B.single_reads_detailed.end() && B.paired_reads_detailed.find(iter->first) == B.paired_reads_detailed.end())
            stcPA.A_reads++;
            
    for(map<string,pair<CBlat::stcAlignment,CBlat::stcAlignment> >::const_iterator iter = A.paired_reads_detailed.begin(); iter != A.paired_reads_detailed.end(); iter++)
        if(B.paired_reads_detailed.find(iter->first) == B.paired_reads_detailed.end())
            stcPA.A_pairs++;
            
    for(map<string, CBlat::stcAlignment>::const_iterator iter = B.single_reads_detailed.begin(); iter != B.single_reads_detailed.end(); iter++)
        if(A.single_reads_detailed.find(iter->first) == A.single_reads_detailed.end() && A.paired_reads_detailed.find(iter->first) == A.paired_reads_detailed.end())
            stcPA.B_reads++;
            
    for(map<string,pair<CBlat::stcAlignment,CBlat::stcAlignment> >::const_iterator iter = B.paired_reads_detailed.begin(); iter != B.paired_reads_detailed.end(); iter++)
        if(A.paired_reads_detailed.find(iter->first) == A.paired_reads_detailed.end())
            stcPA.B_pairs++;
    
    stc_paired_alleles stcReturn;
    stcReturn.A = A;
    stcReturn.B = B;
    stcReturn.reads = stcPA;
            
    return stcReturn;
}
*/

size_t countUniqueAlleleSpecificReads(map<string, CBlat::stcAlignment> reads)
{
    set<string> unique;
    for(map<string, CBlat::stcAlignment>::const_iterator iter = reads.begin(); iter != reads.end(); iter++)
    {
        vector<unsigned int> posses;
        posses.push_back(iter->second.RefId);
        posses.push_back(iter->second.RefStart);
        posses.push_back(iter->second.RefStart+iter->second.AlignLength);
        sort(posses.begin()+1,posses.end());
        ostringstream osr("");
        osr << posses[0] << '_' << posses[1] << '_' << posses[2];
        unique.insert(osr.str());
    }

    return unique.size();
}

// nur die Allelspezifischen reads der angegebenen Exone
size_t countUniqueAlleleSpecificReads(map<string, CBlat::stcAlignment> reads, vector<unsigned int> exon_list)
{
    set<unsigned int> the_refs;
    for(vector<unsigned int>::iterator iter = exon_list.begin(); iter != exon_list.end(); iter++)
        the_refs.insert(*iter);

    for(map<string, CBlat::stcAlignment>::iterator iter = reads.begin(); iter != reads.end(); iter++)
        if( the_refs.find(iter->second.RefId) == the_refs.end() )
            reads.erase(iter--);
    
    return countUniqueAlleleSpecificReads(reads);
}

size_t countUniqueAlleleSpecificPairs(map<string,pair<CBlat::stcAlignment,CBlat::stcAlignment> > reads)
{
    set<string> unique;
    for(map<string,pair<CBlat::stcAlignment,CBlat::stcAlignment> >::const_iterator iter = reads.begin(); iter != reads.end(); iter++)
    {
        vector<unsigned int> possesA;
        vector<unsigned int> possesB;
        possesA.push_back(iter->second.first.RefId);
        possesA.push_back(iter->second.first.RefStart);
        possesA.push_back(iter->second.first.RefStart+iter->second.first.AlignLength);
        sort(possesA.begin()+1,possesA.end());
        ostringstream osrA("");
        osrA << possesA[0] << '_' << possesA[1] << '_' << possesA[2];
        possesB.push_back(iter->second.first.RefId);
        possesB.push_back(iter->second.first.RefStart);
        possesB.push_back(iter->second.first.RefStart+iter->second.first.AlignLength);
        sort(possesB.begin()+1,possesB.end());
        ostringstream osrB("");
        osrB << possesB[0] << '_' << possesB[1] << '_' << possesB[2];
        if(osrA.str().compare(osrB.str()) > 0)
            unique.insert(osrA.str()+osrB.str());
        else
            unique.insert(osrB.str()+osrA.str());
    }
    return unique.size();
}


stc_paired_alleles getAlleleSpecificReads(const stc_mapping_results& A,const stc_mapping_results& B)
{

    map<string, CBlat::stcAlignment> A_reads;
    map<string, CBlat::stcAlignment> B_reads;
    map<string,pair<CBlat::stcAlignment,CBlat::stcAlignment> > A_pairs;
    map<string,pair<CBlat::stcAlignment,CBlat::stcAlignment> > B_pairs;


    for(map<string, CBlat::stcAlignment>::const_iterator iter = A.single_reads_detailed.begin(); iter != A.single_reads_detailed.end(); iter++)
        if(B.single_reads_detailed.find(iter->first) == B.single_reads_detailed.end() && B.paired_reads_detailed.find(iter->first) == B.paired_reads_detailed.end())
            A_reads[iter->first]=iter->second;

    for(map<string,pair<CBlat::stcAlignment,CBlat::stcAlignment> >::const_iterator iter = A.paired_reads_detailed.begin(); iter != A.paired_reads_detailed.end(); iter++)
        if(B.paired_reads_detailed.find(iter->first) == B.paired_reads_detailed.end())
            A_pairs[iter->first]=iter->second;

    for(map<string, CBlat::stcAlignment>::const_iterator iter = B.single_reads_detailed.begin(); iter != B.single_reads_detailed.end(); iter++)
        if(A.single_reads_detailed.find(iter->first) == A.single_reads_detailed.end() && A.paired_reads_detailed.find(iter->first) == A.paired_reads_detailed.end())
            B_reads[iter->first]=iter->second;

    for(map<string,pair<CBlat::stcAlignment,CBlat::stcAlignment> >::const_iterator iter = B.paired_reads_detailed.begin(); iter != B.paired_reads_detailed.end(); iter++)
        if(A.paired_reads_detailed.find(iter->first) == A.paired_reads_detailed.end())
            B_pairs[iter->first]=iter->second;

    stc_reads_per_allele stcPA;

    stcPA.A_reads = countUniqueAlleleSpecificReads(A_reads);
    stcPA.A_pairs = countUniqueAlleleSpecificPairs(A_pairs);
    stcPA.B_reads = countUniqueAlleleSpecificReads(B_reads);
    stcPA.B_pairs = countUniqueAlleleSpecificPairs(B_pairs);


    stc_paired_alleles stcReturn;
    stcReturn.A = A;
    stcReturn.B = B;
    stcReturn.reads = stcPA;

    return stcReturn;
}

stc_mapping_results getReducedMappingResults(const stc_mapping_results& in)
{
    stc_mapping_results out;
    out.alignable_seq_size = in.alignable_seq_size;
    out.aligned_read_count = in.aligned_read_count;
    out.allele             = in.allele;
    out.allele_seq_size    = in.allele_seq_size;
    out.coverage_drop_penalties = in.coverage_drop_penalties;
    out.covered_bases_count= in.covered_bases_count;
    out.exon_boundary_penalties = in.exon_boundary_penalties;
    out.paired_read_count  = in.paired_read_count;
    out.paired_reads_detailed.clear();
    out.single_reads_detailed.clear();
    return out;
}

bool sortAllelePairingResults(const stc_paired_alleles& a,const stc_paired_alleles& b)
{
    /*
    if(a.MWscore > b.MWscore)
        return true;
    else if(a.MWscore < b.MWscore)
        return false;
    */

    // F1 score !!!!!!!!!!!
    if(a.Fscore > b.Fscore)
        return true;
    else if(a.Fscore < b.Fscore)
        return false;

    // ##############################
    //cout << a.A.allele << '\t' << a.B.allele << " vs " << b.A.allele << '\t' << b.B.allele << endl;
    /*if(a.A.covered_bases_count/a.A.alignable_seq_size+a.B.covered_bases_count/a.B.alignable_seq_size > b.A.covered_bases_count/b.A.alignable_seq_size+b.B.covered_bases_count/b.B.alignable_seq_size)
        return true;
    else if(a.A.covered_bases_count/a.A.alignable_seq_size+a.B.covered_bases_count/a.B.alignable_seq_size > b.A.covered_bases_count/b.A.alignable_seq_size+b.B.covered_bases_count/b.B.alignable_seq_size)
        return false;*/
    if(a.reads.A_pairs+a.reads.B_pairs > b.reads.A_pairs+b.reads.B_pairs)
        return true;
    else if(a.reads.A_pairs+a.reads.B_pairs < b.reads.A_pairs+b.reads.B_pairs)
        return false;
    if(a.reads.A_reads+a.reads.B_reads > b.reads.A_reads+b.reads.B_reads)
        return true;
    else if(a.reads.A_reads+a.reads.B_reads < b.reads.A_reads+b.reads.B_reads)
        return false;
    return false;
}

/*
void calculateMWScore(vector<stc_paired_alleles>& values)
{
    for(vector<stc_paired_alleles>::iterator iter = values.begin(); iter != values.end(); iter++)
    {
        double A_reads_per_base = (double)iter->reads.A_reads/iter->A.alignable_seq_size;
        double B_reads_per_base = (double)iter->reads.B_reads/iter->B.alignable_seq_size;
        iter->read_per_base = A_reads_per_base+B_reads_per_base;

        double req = 0.0f;
        if(A_reads_per_base != 0 && B_reads_per_base != 0)
            req = 1-( (max(A_reads_per_base,B_reads_per_base)/min(A_reads_per_base,B_reads_per_base))-1 );
        iter->read_equality = req;

        double act_pairs_per_Read = 0.0f;
        if( iter->reads.A_reads+iter->reads.B_reads != 0 )
            act_pairs_per_Read = (double)(iter->reads.A_pairs+iter->reads.B_pairs)/(iter->reads.A_reads+iter->reads.B_reads);

        double area_under_curve = (iter->A.area_under_curve_error+iter->B.area_under_curve_error)/2;

        iter->Fscore = iter->read_per_base*iter->read_equality*area_under_curve/(iter->read_per_base+iter->read_equality+area_under_curve);
    }
}
*/

void calculateMWScore(vector<stc_paired_alleles>& values)
{
    set<double> reads_per_base;
    set<double> pairs_per_read;
    set<double> reads_equality;
    set<double> area_under_curve;
    set<size_t> bases;
    // set<size_t> unalignable_bases; // Hier werden die basen gezählt die wegen der kürze der exone nicht gematch werden können. Sie gelten hier als unsicherheitsfaktor

    for(vector<stc_paired_alleles>::iterator iter = values.begin(); iter != values.end(); iter++)
    {
        double A_reads_per_base =   (double)iter->reads.A_reads/iter->A.alignable_seq_size;
        double B_reads_per_base =   (double)iter->reads.B_reads/iter->B.alignable_seq_size;
        double sum_reads_per_base = ((double)(iter->reads.A_reads+iter->reads.B_reads))/(iter->A.alignable_seq_size+iter->B.alignable_seq_size);
        iter->read_per_base = *reads_per_base.insert(sum_reads_per_base).first;
        
        // double req = 0.0f;
        // if(A_reads_per_base != 0 && B_reads_per_base != 0)
        //    req = 1-( (max(A_reads_per_base,B_reads_per_base)/min(A_reads_per_base,B_reads_per_base))-1 );
        double req = fabs(A_reads_per_base-B_reads_per_base);
        iter->read_equality = *reads_equality.insert(req).first;

        double act_pairs_per_Read = 0.0f;
        if( iter->reads.A_reads+iter->reads.B_reads != 0 )
            act_pairs_per_Read = (double)(iter->reads.A_pairs+iter->reads.B_pairs)/(iter->reads.A_reads+iter->reads.B_reads);
        iter->pairs_per_read = *pairs_per_read.insert( act_pairs_per_Read ).first;

        bases.insert(iter->A.covered_bases_count+iter->B.covered_bases_count);

        double aauc = (iter->A.area_under_curve_error*iter->A.alignable_seq_size+iter->B.area_under_curve_error*iter->B.alignable_seq_size)/(iter->A.alignable_seq_size+iter->B.alignable_seq_size);
        iter->area_under_curve_error = *area_under_curve.insert(aauc ).first;
    }

    double max_reads_per_base = *reads_per_base.rbegin();
    double min_reads_per_base = *reads_per_base.begin();
    // Achtung
    min_reads_per_base = 0;
    double div_reads_per_base = max_reads_per_base - min_reads_per_base;

    double max_reads_equality = 1.0f;
    double min_reads_equality = *reads_equality.begin();
    // Achtung
    // max_reads_equality = max_reads_equality*100;
    double div_reads_equality = max_reads_equality - min_reads_equality;


    double max_pairs_per_read = *pairs_per_read.rbegin();
    double min_pairs_per_read = *pairs_per_read.begin();
    // Achtung
    min_pairs_per_read = 0.0f;
    double div_pairs_per_read = max_pairs_per_read - min_pairs_per_read;

    size_t max_bases = *bases.rbegin();
    size_t min_bases = *bases.begin();
    // Achtung
    min_bases = 0;
    size_t div_bases = max_bases - min_bases;

    double max_area_under_curve = *area_under_curve.rbegin();
    // Achtung
    max_area_under_curve = 100.0f;
    double min_area_under_curve = *area_under_curve.begin();
    double div_area_under_curve = max_area_under_curve-min_area_under_curve;

    double f_score = 0.0f;
    for(vector<stc_paired_alleles>::iterator iter = values.begin(); iter != values.end(); iter++)
    {
        double A_reads_per_base = (double)iter->reads.A_reads/iter->A.alignable_seq_size;
        double B_reads_per_base = (double)iter->reads.B_reads/iter->B.alignable_seq_size;
        double reads_per_base = ((double)(iter->reads.A_reads+iter->reads.B_reads))/(iter->A.alignable_seq_size+iter->B.alignable_seq_size);
        double val_reads_per_base = 1.0f;
        if(div_reads_per_base != 0.0f)
            val_reads_per_base = ( reads_per_base - min_reads_per_base ) / div_reads_per_base;

        // double req = 0.0f;
        // if(A_reads_per_base != 0 && B_reads_per_base != 0)
        //    req = 1-( (max(A_reads_per_base,B_reads_per_base)/min(A_reads_per_base,B_reads_per_base))-1 );
        //double val_reads_equality = 1.0f;
        //if(div_reads_equality != 0.0f)
        //    val_reads_equality = ( req - min_reads_equality ) / div_reads_equality;
        double val_reads_equality = 0.0f;
        double req = fabs(A_reads_per_base-B_reads_per_base);
        if(div_reads_equality != 0.0f)
            val_reads_equality = 1 -(( req - min_reads_equality ) / div_reads_equality);

        double val_bases = 1.0f;
        if(div_bases != 0.0f)
            val_bases = (double)(iter->A.covered_bases_count+iter->B.covered_bases_count-min_bases)/div_bases;

        double act_pairs_per_Read = 0.0f;
        if( iter->reads.A_reads+iter->reads.B_reads != 0 )
            act_pairs_per_Read = (double)(iter->reads.A_pairs+iter->reads.B_pairs)/(iter->reads.A_reads+iter->reads.B_reads);
        double val_pairs_per_Read = 1.0f;
        if(div_pairs_per_read != 0.0f)
            val_pairs_per_Read = (act_pairs_per_Read - min_pairs_per_read) / div_pairs_per_read;

        double act_area_under_curve = (iter->A.area_under_curve_error*iter->A.alignable_seq_size+iter->B.area_under_curve_error*iter->B.alignable_seq_size)/(iter->A.alignable_seq_size+iter->B.alignable_seq_size);
        double val_area_under_curve = 1.0f;
        if(div_area_under_curve != 0.0f)
            val_area_under_curve = 1- ((act_area_under_curve - min_area_under_curve) / div_area_under_curve);

        //cout << iter->A.allele << "\t" << iter->B.allele << endl;
        iter->Fscore = (2 + 1 + 10 + 10 + 1)  * val_reads_per_base*val_reads_equality*val_bases*val_pairs_per_Read*val_area_under_curve/(2*val_reads_per_base+1*val_reads_equality+10*val_bases+10*val_pairs_per_Read+1*val_area_under_curve);
        //iter->Fscore = val_reads_per_base*val_reads_equality*val_pairs_per_Read*val_area_under_curve/(val_reads_per_base+val_reads_equality+val_pairs_per_Read+val_area_under_curve);
    }

}

// Dies ist die neue Version, die nur Exone Vergleicht, die lang genug fürs alignment sind
void removeShortCDNAbias(vector<stc_mapping_results>& best_cDNA_hits,map<string,string>& imgtdb, bool prefer_longer_annotation)
{
    for(vector<stc_mapping_results>::iterator iter = best_cDNA_hits.begin(); iter != best_cDNA_hits.end(); iter++)
    {
        CHLAAlignment alignment(imgtdb[iter->allele]);
        for(vector<stc_mapping_results>::iterator iterA = best_cDNA_hits.begin(); iterA != best_cDNA_hits.end(); iterA++)
        {
            vector<stc_mapping_results>::iterator iterB = iterA;
            for(iterB++; iterB != best_cDNA_hits.end(); iterB++)
            {
                CHLAAlignment alignmentA(imgtdb[iterA->allele]);
                CHLAAlignment alignmentB(imgtdb[iterB->allele]);
                if( ( iter->allele != iterA->allele && iter->allele != iterB->allele) &&
                    ( alignment.getAlignableExonCount() <= alignmentA.getAlignableExonCount() || alignment.getAlignableExonCount() <= alignmentB.getAlignableExonCount()) )
                {
                    int count_euqality = 0;
                    for(int i = 0; i < alignment.m_exon_sequences.size(); i++)
                    {
                        for(int j = 0; j < alignmentA.m_exon_sequences.size(); j++)
                        {
                            if(alignment.isAlignableExon(i) && alignment.m_exon_sequences[i].compare(alignmentA.m_exon_sequences[j]) == 0)
                            {
                                count_euqality++;
                                goto compare_next;
                            }
                        }
                        for(int j = 0; j < alignmentB.m_exon_sequences.size(); j++)
                        {
                            if(alignment.isAlignableExon(i) && alignment.m_exon_sequences[i].compare(alignmentB.m_exon_sequences[j]) == 0)
                            {
                                count_euqality++;
                                goto compare_next;
                            }
                        }
                        compare_next:;
                    }
                    if(count_euqality == alignment.getAlignableExonCount())
                    {
                        if( !prefer_longer_annotation ||
                            (prefer_longer_annotation && isTheSameAtTwoDigit(iter->allele,iterA->allele) && !isAncientAnnotation(iter->allele,iterA->allele)) ||
                            (prefer_longer_annotation && isTheSameAtTwoDigit(iter->allele,iterB->allele) && !isAncientAnnotation(iter->allele,iterB->allele))
                           )
                        {
                            cout << "erase " << iter->allele << " vs " << iterA->allele << " - " << iterB->allele << endl;
                            best_cDNA_hits.erase(iter--);
                            goto jump_point_after_erase;
                        }
                    }
                }
            }
        }
        jump_point_after_erase:;
    }
}

// Dies ist die alte Version, die alle Exone vergleicht
/*
void removeShortCDNAbias(vector<stc_mapping_results>& best_cDNA_hits,map<string,string>& imgtdb, bool prefer_longer_annotation)
{
    for(vector<stc_mapping_results>::iterator iter = best_cDNA_hits.begin(); iter != best_cDNA_hits.end(); iter++)
    {
        CHLAAlignment alignment(imgtdb[iter->allele]);
        for(vector<stc_mapping_results>::iterator iterA = best_cDNA_hits.begin(); iterA != best_cDNA_hits.end(); iterA++)
        {
            vector<stc_mapping_results>::iterator iterB = iterA;
            for(iterB++; iterB != best_cDNA_hits.end(); iterB++)
            {
                CHLAAlignment alignmentA(imgtdb[iterA->allele]);
                CHLAAlignment alignmentB(imgtdb[iterB->allele]);
                if( ( iter->allele != iterA->allele && iter->allele != iterB->allele) &&
                    ( alignment.m_exon_sequences.size() <= alignmentA.m_exon_sequences.size() || alignment.m_exon_sequences.size() <= alignmentB.m_exon_sequences.size()) )
                {
                    int count_euqality = 0;
                    for(int i = 0; i < alignment.m_exon_sequences.size(); i++)
                    {
                        for(int j = 0; j < alignmentA.m_exon_sequences.size(); j++)
                        {
                            if(alignment.m_exon_sequences[i].compare(alignmentA.m_exon_sequences[j]) == 0)
                            {
                                count_euqality++;
                                goto compare_next;
                            }
                        }
                        for(int j = 0; j < alignmentB.m_exon_sequences.size(); j++)
                        {
                            if(alignment.m_exon_sequences[i].compare(alignmentB.m_exon_sequences[j]) == 0)
                            {
                                count_euqality++;
                                goto compare_next;
                            }
                        }
                        compare_next:;
                    }
                    if(count_euqality == alignment.m_exon_sequences.size())
                    {
                        if( !prefer_longer_annotation ||
                            (prefer_longer_annotation && isTheSameAtTwoDigit(iter->allele,iterA->allele) && !isAncientAnnotation(iter->allele,iterA->allele)) ||
                            (prefer_longer_annotation && isTheSameAtTwoDigit(iter->allele,iterB->allele) && !isAncientAnnotation(iter->allele,iterB->allele))
                           )
                        {
                            cout << "erase " << iter->allele << " vs " << iterA->allele << " - " << iterB->allele << endl;
                            best_cDNA_hits.erase(iter--);
                            goto jump_point_after_erase;
                        }
                    }
                }
            }
        }
        jump_point_after_erase:;
    }
}
*/

// Vergleiche Namensgebung. Ist A ursprünglicher als b? Mehr Felder ist Ursprünglicher. Bei gleichen Feldern ist von vorn nach hinten die kleinere Nummer ursprünglicher
bool isAncientAnnotation(const string& a, const string& b)
{
    string adigits = Tokenize(a,"*")[1];
    string bdigits = Tokenize(b,"*")[1];
    vector<string> ad = Tokenize(adigits,":");
    vector<string> bd = Tokenize(bdigits,":");

    if(ad.size() > bd.size())
        return true;
    if(ad.size() < bd.size())
        return false;
    for(int i = 0; i < ad.size(); i++)
    {
        if(atoi(ad[i].c_str()) < atoi(bd[i].c_str()))
            return true;
        if(atoi(ad[i].c_str()) > atoi(bd[i].c_str()))
            return false;
    }
    return false;
}

bool isTheSameAtTwoDigit(const string& a, const string& b)
{
    string adigits = Tokenize(a,"*")[1];
    string bdigits = Tokenize(b,"*")[1];
    vector<string> ad = Tokenize(adigits,":");
    vector<string> bd = Tokenize(bdigits,":");


    if(atoi(ad[0].c_str()) == atoi(bd[0].c_str()))
        return true;
    return false;

}

void bringUpTheMostAncientAllleleCombination(vector<stc_paired_alleles>& allele_gt,double f_score_within)
{
    vector<stc_paired_alleles>::iterator iter_most_ancient = allele_gt.begin();

    double max_fscore = iter_most_ancient->Fscore;
    double min_fscore = iter_most_ancient->Fscore;

    for(int i = 1; i < allele_gt.size(); i++)
    {
        if(allele_gt[i].Fscore < min_fscore)
            min_fscore = allele_gt[i].Fscore;
        if(allele_gt[i].Fscore > max_fscore)
            max_fscore = allele_gt[i].Fscore;
    }
    for(vector<stc_paired_alleles>::iterator iter = allele_gt.begin(); iter != allele_gt.end() && iter->Fscore >= max_fscore*f_score_within; iter++)
    {
        set<string> best_pair;
        best_pair.insert(iter_most_ancient->A.allele);
        best_pair.insert(iter_most_ancient->B.allele);
        string best_A = *best_pair.begin();
        string best_B = *best_pair.rbegin();

        set<string> act_pair;
        act_pair.insert(iter->A.allele);
        act_pair.insert(iter->B.allele);
        string act_A = *act_pair.begin();
        string act_B = *act_pair.rbegin();

        if(iter->area_under_curve_error > iter_most_ancient->area_under_curve_error)
            continue;

        if(
               ( act_A == best_A && isAncientAnnotation(act_B,best_B) ) ||
               ( act_B == best_B && isAncientAnnotation(act_A,best_A) ) ||
               ( isAncientAnnotation(act_A,best_A) && isAncientAnnotation(act_B,best_B)  ) ||
               ( act_A == best_B && isAncientAnnotation(act_B,best_A) ) ||
               ( act_B == best_A && isAncientAnnotation(act_A,best_B) ) ||
               ( isAncientAnnotation(act_A,best_B) && isAncientAnnotation(act_B,best_A)  )
          )
        {
            //read_per_base   pairs_per_read  read_equality   area_under_curve_error    Fscore
            cout << "bring_up_most_ancient\t" << iter_most_ancient->A.allele << '\t' << iter_most_ancient->B.allele << '\t'
                 << iter_most_ancient->read_per_base  << '\t'
                 << iter_most_ancient->pairs_per_read  << '\t'
                 << iter_most_ancient->read_equality << '\t'
                 << iter_most_ancient->area_under_curve_error  << '\t'
                 << iter_most_ancient->Fscore << '\t'
                 << iter->A.allele << '\t' << iter->B.allele << '\t'
                 << iter->read_per_base  << '\t'
                 << iter->pairs_per_read  << '\t'
                 << iter->read_equality << '\t'
                 << iter->area_under_curve_error  << '\t'
                 << iter->Fscore << endl;

            iter_most_ancient = iter;
        }
    }
    stc_paired_alleles hlp = *iter_most_ancient;
    allele_gt.erase(iter_most_ancient);
    allele_gt.insert(allele_gt.begin(),hlp);
}

void bringUpTheMostAncientHomozygousAllele(vector<stc_paired_alleles>& allele_gt,double inequality_factor)
{

    vector<stc_paired_alleles>::iterator iter_most_ancient = allele_gt.begin();

    if(iter_most_ancient->A.allele == iter_most_ancient->B.allele)
        return;

    //cout << iter_most_ancient->A.allele << "\t" << iter_most_ancient->A.aligned_read_count << '(' << iter_most_ancient->reads.A_reads << ')' << "\t" << iter_most_ancient->A.alignable_seq_size << " - "
    //     << iter_most_ancient->B.allele << "\t" << iter_most_ancient->B.aligned_read_count << '(' << iter_most_ancient->reads.B_reads << ')' << "\t" << iter_most_ancient->B.alignable_seq_size << endl;

    //float reads_per_base_A = (float)iter_most_ancient->A.aligned_read_count/iter_most_ancient->A.alignable_seq_size;
    //float reads_per_base_B = (float)iter_most_ancient->B.aligned_read_count/iter_most_ancient->B.alignable_seq_size;
    float reads_per_base_A = ((float)iter_most_ancient->reads.A_reads)/((float)iter_most_ancient->A.alignable_seq_size);
    float reads_per_base_B = ((float)iter_most_ancient->reads.B_reads)/((float)iter_most_ancient->B.alignable_seq_size);

    //cout << reads_per_base_A << "\t" << reads_per_base_B << endl;

    if(
           isTheSameAtTwoDigit(iter_most_ancient->A.allele, iter_most_ancient->B.allele) &&
        ( reads_per_base_A == 0.0f || reads_per_base_B == 0.0f || reads_per_base_A/reads_per_base_B >= inequality_factor || reads_per_base_B/reads_per_base_A >= inequality_factor)
      )
    {
        string the_ancient_one = iter_most_ancient->A.allele;
        if(!isAncientAnnotation(the_ancient_one,iter_most_ancient->B.allele))
            the_ancient_one = iter_most_ancient->A.allele;
        for(vector<stc_paired_alleles>::iterator iter = allele_gt.begin(); iter != allele_gt.end() ; iter++)
            if(iter->A.allele == iter->B.allele && iter->A.allele == the_ancient_one)
                iter_most_ancient = iter;
    }
    else
        return;

    stc_paired_alleles hlp = *iter_most_ancient;
    allele_gt.erase(iter_most_ancient);
    allele_gt.insert(allele_gt.begin(),hlp);
}

vector<string> removeAmbiguityCandidates(vector<stc_mapping_results>& best_cDNA_hits,CImgtAmbiguities& imgt)
{
    vector<string> vctRet;

    vector<string> the_alleles;
    for(vector<stc_mapping_results>::iterator iter = best_cDNA_hits.begin(); iter != best_cDNA_hits.end(); iter++)
        the_alleles.push_back(iter->allele);
    for(vector<stc_mapping_results>::iterator iter = best_cDNA_hits.begin(); iter != best_cDNA_hits.end(); iter++)
        if(imgt.isAmbiguityOfBetterGAllele(iter->allele,the_alleles))
        {
            vctRet.push_back(iter->allele);
            best_cDNA_hits.erase(iter--);
        }
    return vctRet;
}


void eraseRedundantSequences(map<string,string>& idb)
{
    try
    {
        map<string,string>  uniqueSequences;
        for(map<string,string>::iterator iter = idb.begin(); iter != idb.end(); iter++)
        {
            CHLAAlignment al(iter->second);
            map<string,string>::iterator iter_bekannt =  uniqueSequences.find(al.getAlignableSequence());
            if(iter_bekannt != uniqueSequences.end())
            {
                const string& old_allele = iter_bekannt->second;
                const string& new_allele = iter->first;
                if(isAncientAnnotation(old_allele,new_allele))
                {
                    idb.erase(iter--);
                    cout << old_allele << " replaces " << new_allele << endl;
                }
                else
                {
                    iter = idb.find(old_allele);
                    iter_bekannt->second = new_allele;
                    idb.erase(iter--);
                    cout << new_allele << " replaces " << old_allele << endl;
                }
                // cout << iter->first << endl;
            }
            else
                uniqueSequences[al.getAlignableSequence()]=iter->first;
        }
    }
    catch(...)
    {
        cerr << "Unexpected error in \"void eraseRedundantSequences(CImgtDb& idb)\". Very likely bad iterator handling ..." << endl;
        exit(EXIT_FAILURE);
    }
}

string getAlleleName(const string& allele,unsigned int resolution)
{
    ostringstream strReturn("");
    vector<string> parsed = Tokenize(allele,":");
    strReturn << parsed[0];
    int i = 1;
    while(i < resolution && i < parsed.size())
    {
        strReturn << ':' << parsed[i++];
        //cout << strReturn.str() << endl;
        
    }
    
    return strReturn.str();
}

vector<string> getAlternativeAlleles(const vector<stc_paired_alleles>& val)
{
    vector<string> vctReturn;
    if(val.size() == 0)
        return vctReturn;
    set<string> setUnique;

    stc_paired_alleles ref = val[0];
    
    for(int i = 1; i < val.size(); i++)
    {
        if(val[i].A.allele != ref.A.allele && val[i].A.allele != ref.B.allele)
            setUnique.insert(val[i].A.allele);
        if(val[i].B.allele != ref.A.allele && val[i].B.allele != ref.B.allele)
            setUnique.insert(val[i].B.allele);
    }
    for(set<string>::iterator iter = setUnique.begin(); iter != setUnique.end(); iter++)
        vctReturn.push_back(*iter);
    return vctReturn;
}

string getAllelesAstring(const vector<string>& val, unsigned int resolution)
{
    set<string> setUnique;
    for(vector<string>::const_iterator iter = val.begin(); iter != val.end(); iter++)
        setUnique.insert(getAlleleName(*iter,resolution));
    ostringstream strRet("");
    for(set<string>::const_iterator iter = setUnique.begin(); iter != setUnique.end(); iter++)
    {
        if(strRet.str().size() > 0)
            strRet << ',';
        strRet << *iter;
    }
    return strRet.str();
}

void removeAllelesWhichAreEquivalent(const string& allel,vector<string>& alleles, unsigned int resolution)
{
    for(vector<string>::iterator iter = alleles.begin(); iter != alleles.end(); iter++)
    {
        if(getAlleleName(*iter,resolution) == getAlleleName(allel,resolution))
            alleles.erase(iter);
    }
}

