/*
 * File:   CImgtDb.cpp
 * Author: mwittig
 *
 * Created on December 6, 2011, 9:01 AM
 */
#include <stdlib.h>

#include <string>
#include <iostream>
#include <sstream>
#include <ostream>
#include <iomanip>
#include <stdint.h>
#include <fstream>
#include <vector>
#include <list>
#include <map>
#include <set>
#include <string.h>
#include <algorithm>
#include <iterator>
#include <numeric>
#include <exception>

#include <boost/regex.hpp>
#include <boost/regex.h>

#include "CMWException.h"
#include "CImgtDb.h"

#define BOOST_ALL_NO_LIB

CImgtDb::CImgtDb() {
}

CImgtDb::CImgtDb(const char* filename)
{
    the_alignments = getAlignmentAsMap(filename,"plus");
}

CImgtDb::CImgtDb(const char* filename,const char* filename_bias)
{
    the_alignments = getAlignmentAsMap(filename,"plus");
    removeSequencesFromBiasFile(filename_bias);
}

CImgtDb::CImgtDb(const CImgtDb& orig)
{
    the_alignments = orig.the_alignments;
}

CImgtDb::~CImgtDb() {
}


/**
 * Start for *_gen.txt
 */
/*
bool CImgtDb::findFaStart(ifstream& the_file)
{
    string strLine;
    while(getline(the_file,strLine) && strLine.find("gDNA") == string::npos);
    if(getline(the_file,strLine))
        return true;
    return false;
}
//*/

/**
 * Start for *_nuc.txt
 */
//*
bool CImgtDb::findFaStart(ifstream& the_file)
{
    string strLine;
    while(getline(the_file,strLine) && strLine.find("cDNA") == string::npos && strLine.find("gDNA") == string::npos);
    while(getline(the_file,strLine))
    {
        boost::regex expression(" +\\|");
        boost::cmatch what;
        if(boost::regex_match(strLine.c_str(), what, expression))
            return true;
    }
    return false;
}
//*/

pair<string,string> CImgtDb::operator[](const string& val)
{
    map<string,string>::const_iterator iter = the_alignments.mapMultipleAlignment.find(val);
    if(iter == the_alignments.mapMultipleAlignment.end())
        return pair<string,string>("","");
    string strSeq = erase_characters("*.", iter->second);
    while(strSeq[0] == '|')
        strSeq.erase(strSeq.begin());
    while(strSeq[strSeq.size()-1] == '|')
        strSeq.erase(--strSeq.end());

    return pair<string,string>(iter->first,strSeq);
}

pair<string,string> CImgtDb::getUnmanipulatedSequence(const string& val)
{
    map<string,string>::const_iterator iter = the_alignments.mapMultipleAlignment.find(val);
    if(iter == the_alignments.mapMultipleAlignment.end())
        return pair<string,string>("","");
    return pair<string,string>(iter->first,iter->second);
}

pair<string,string> CImgtDb::readMultiAlignLine(ifstream& the_file)
{
    string strLine;
    getline(the_file,strLine);
    pair<string,string> pairReturn("","");
    //cout << strLine << endl;

    list<string> l;

    int split_count = boost::regex_split(std::back_inserter(l), strLine);

    int i = 0;
    if(l.size())
    {
        pairReturn.first=*(l.begin());
        l.pop_front();
        while(l.size())
        {
            pairReturn.second+=*(l.begin());
            //cout << setw(10) << right << i++ << '\t' << *(l.begin()) << endl;
            l.pop_front();
        }
    }
    //cout << setw(10) << right << pairReturn.first << '\t' << pairReturn.second << endl;
    return pairReturn;
}

string CImgtDb::getFaSequence(string strTmp,unsigned int lineBreaks)
{
    if(lineBreaks==0)
        return strTmp;
    ostringstream strReturn("");
    unsigned int counter = 0;
    for(string::const_iterator iter = strTmp.begin(); iter != strTmp.end(); iter++)
    {
        strReturn << *iter;
        if(++counter % lineBreaks == 0)
            strReturn << endl;
    }
    return strReturn.str();
}

CImgtDb::multiSeqAlignment CImgtDb::getAlignmentAsMap(const char* filename,const char* strand)
{
    
    try
    {
        multiSeqAlignment mapReturn;
        ifstream in(filename);
        if(!in)
            throw(string("Can't open ")+filename);

        mapReturn.strTemplateAllele="";
        mapReturn.strTemplateSequence="";
        if(findFaStart(in))
        {
            do
            {
                pair<string,string> strLine;
                strLine = readMultiAlignLine(in);
                if(mapReturn.strTemplateAllele.compare("")==0)
                    mapReturn.strTemplateAllele = strLine.first;
                else if(mapReturn.strTemplateAllele.compare(strLine.first) != 0)
                    throw("Template sequence changed within multiple alignment file. Expected "+mapReturn.strTemplateAllele+" and got "+strLine.first+".");
                mapReturn.strTemplateSequence+= strLine.second;
                //cout << mapReturn.strTemplateAllele << "|\t|" << mapReturn.strTemplateSequence << endl;
                strLine = readMultiAlignLine(in);
                while(strLine.second != "")
                {
                    if(mapReturn.mapMultipleAlignment.find(strLine.first) == mapReturn.mapMultipleAlignment.end())
                        mapReturn.mapMultipleAlignment[strLine.first]=strLine.second;
                    else
                        mapReturn.mapMultipleAlignment[strLine.first]+=strLine.second;
                    //cout << strLine << endl;
                    strLine = readMultiAlignLine(in);
                }
            }while(findFaStart(in));
        }
        else
            throw(string("Start not found"));
        in.close();
        for(size_t idx = 0; idx < mapReturn.strTemplateSequence.size(); idx++)
        {
            for(map<string,string>::iterator iter = mapReturn.mapMultipleAlignment.begin(); iter != mapReturn.mapMultipleAlignment.end(); iter++)
            {
                switch(iter->second[idx])
                {
                    case '-':
                       iter->second[idx] =  mapReturn.strTemplateSequence[idx];
                }
            }
        }

        if(string("minus").compare(strand)== 0)
        {
            for(map<string,string>::iterator iter = mapReturn.mapMultipleAlignment.begin(); iter != mapReturn.mapMultipleAlignment.end(); iter++)
                iter->second = GetComplSequence(iter->second);
            mapReturn.strTemplateSequence = GetComplSequence(mapReturn.strTemplateSequence);
        }
        else if(string("plus").compare(strand) != 0)
            throw(string("No strand specified. Your value is \"")+strand+"\", but only plus or minus are allowed ...");

        mapReturn.mapMultipleAlignment[mapReturn.strTemplateAllele]=mapReturn.strTemplateSequence;
        //cout << mapReturn.strTemplateAllele << "!!!" << endl;
        /*
        cout << setw(20) << left << strTemplateAllele << '\t' << strTemplateSequence << endl;
        for(map<string,string>::iterator iter = mapMultipleAlignment.begin(); iter != mapMultipleAlignment.end(); iter++ )
            cout << setw(20) << left << iter->first << '\t' << iter->second << endl;
        */
        //cout << mapReturn.strTemplateSequence << endl;
        return mapReturn;
    }
    catch(string& err)
    {
        cerr << err << endl;
        exit(EXIT_FAILURE);
    }
    catch(...)
    {
        cerr << "Unknown error occured" << endl;
        exit(EXIT_FAILURE);
    }
    
}

string CImgtDb::erase_characters (string rep, string in)
{
  int pos;
  while (true)
  {
    pos = in.find_first_of(rep);
    if (pos == -1)
      break;
    else
      in.erase(pos, 1);
  }
  return in;
}




char CImgtDb::GetComplNuc(const char& source)
{
	switch(source)
	{
	case 'A':
		return 'T';
	case 'a':
		return 't';
	case 'C':
		return 'G';
	case 'c':
		return 'g';
	case 'G':
		return 'C';
	case 'g':
		return 'c';
	case 'T':
	case 'U':
		return 'A';
	case 't':
	case 'u':
		return 'a';
	case 'M':
		return 'K';
	case 'm':
		return 'k';
	case 'R':
		return 'Y';
	case 'r':
		return 'y';
	case 'Y':
		return 'R';
	case 'y':
		return 'r';
	case 'K':
		return 'M';
	case 'k':
		return 'm';
	case 'V':
		return 'B';
	case 'v':
		return 'b';
	case 'H':
		return 'D';
	case 'h':
		return 'd';
	case 'D':
		return 'H';
	case 'd':
		return 'h';
	case 'B':
		return 'V';
	case 'b':
		return 'v';
	case '(':
         return ')';
    case ')':
         return '(';
	case '[':
         return ']';
    case ']':
         return '[';
	}
	return source;
}

string CImgtDb::GetComplSequence(string& strSequence)
{
	try
    {
        int intPos = 0;
    	size_t intSize = strSequence.length();
    	char* buffer=(char*)malloc(intSize+1);
        if(!buffer)
            throw(string("can't allocate enough memory."));
        buffer[intSize]=0;
        while(--intSize != -1)
        {
            buffer[intPos++]=GetComplNuc(strSequence.at(intSize));
        }
       	string strReturn(buffer);
        free(buffer);
        return strReturn;
    }
    catch(string strError)
    {
        throw(string("error in CMyTools::GetSeqFromFile, ")+strError);
    }
    catch(...)
    {
        throw("unknown error in CMyTools::GetSeqFromFile\n");
    }

}


char CImgtDb::getOneLetterNucleotide(set<char>& the_nucs)
{
    string strLine("");
    for(set<char>::const_iterator iter = the_nucs.begin(); iter != the_nucs.end(); iter++)
        strLine+=*iter;
    if(strLine.compare("A")==0)
        return 'A';
    if(strLine.compare("C")==0)
        return 'C';
    if(strLine.compare("G")==0)
        return 'G';
    if(strLine.compare("T")==0)
        return 'T';
    if(strLine.compare("AG")==0)
        return 'R';
    if(strLine.compare("CT")==0)
        return 'Y';
    if(strLine.compare("AC")==0)
        return 'M';
    if(strLine.compare("GT")==0)
        return 'K';
    if(strLine.compare("CG")==0)
        return 'S';
    if(strLine.compare("AT")==0)
        return 'W';
    if(strLine.compare("ACT")==0)
        return 'H';
    if(strLine.compare("CGT")==0)
        return 'B';
    if(strLine.compare("ACG")==0)
        return 'V';
    if(strLine.compare("AGT")==0)
        return 'D';
    return 'N';

}


bool CImgtDb::isRegularNucleotide(const char val)
{
    switch(val)
    {
        case 'A':
        case 'C':
        case 'G':
        case 'T':
        case 'a':
        case 'c':
        case 'g':
        case 't':
            return true;
    }
    return false;

}

map<string, set<string> > CImgtDb::getUniqueSequences(bool remove_special_characters)
{
    map<string, set<string> > mapReturn;
    for(map<string,string>::const_iterator iter = the_alignments.mapMultipleAlignment.begin(); iter != the_alignments.mapMultipleAlignment.end(); iter++)
    {
        string actSeq = iter->second;
        if(remove_special_characters)
            actSeq = erase_characters("*.|",actSeq);
        map<string, set<string> >::iterator iterExists = mapReturn.find(actSeq);
        if(iterExists == mapReturn.end())
            mapReturn[actSeq].insert(iter->first);
        else
            iterExists->second.insert(iter->first);
    }
    return mapReturn;
}

void CImgtDb::removeSequencesFromBiasFile(const char* filename)
{
    ifstream in(filename);
    if(!in)
        throw(string("Can't open ")+filename);
    string strLine;

    while(getline(in,strLine))
    {
        map<string,string>::iterator iter = the_alignments.mapMultipleAlignment.find(strLine);
        if(iter != the_alignments.mapMultipleAlignment.end())
        {
            // cerr << "removed " << iter->first << endl;
            the_alignments.mapMultipleAlignment.erase(iter);
        }
    }

    in.close();
}
