/*************************************************
 * 
    CMyTools , is a collection of frequently use functions
    Copyright (C) [2006]  [Michael Wittig] (m.wittig@mucosa.de  "www.ikmb.uni-kiel.de")

    This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License 
    as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; 
    if not, write to the Free Software Foundation, 
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

***************************************************/

#include <string>
#include <string.h>
#include <assert.h>
#include <vector>
#include <iostream>
#include <sstream>
#include <algorithm>
#include <dirent.h>
#include <stdlib.h>
#include "MyTools.h"

using namespace std;

CMyTools::CMyTools()
{
}

CMyTools::~CMyTools()
{
}

vector<string> 	CMyTools::GetParsedLine(string line)
{
    return GetParsedLine(line,"\t");			
}

vector<string> CMyTools::GetParsedLine(string str,const char* delim)
{
    vector<string>  results;
    string::size_type  delimSize=string(delim).length();
    string::size_type  delimAt;
    string::size_type  idx = 0;
    while( (delimAt = str.find(delim,idx)) != string::npos )
    {
        if(delimAt != string::npos)
            results.push_back(str.substr(idx,delimAt-idx));
        idx = delimAt+delimSize;
     }
    if(str.length() > idx)
        results.push_back(str.substr(idx));
    return results;
}


// Bitte bei Gelegenheit optimieren !!!!!!!!!!!!
string	CMyTools::ReplaceSubstr(std::string result, const std::string& replaceWhat,  const std::string& replaceWithWhat)
{
  while(1)
  {
    const int pos = result.find(replaceWhat);
    if (pos==-1) break;
    result.replace(pos,replaceWhat.size(),replaceWithWhat);
  }
  return result;
}



vector<string> CMyTools::GetDirectoryListing(const char* path)
{
    DIR *dirp;
    struct dirent *entry;
    vector<string> strReturn;
    
    if(dirp = opendir(path))
    {
        while(entry = readdir(dirp))
            strReturn.push_back(entry->d_name);
        closedir(dirp);
    }   
    return strReturn;
}

vector<string> CMyTools::GetDirectoryListing(const char* path,const char* contains)
{
    vector<string> strReturn;
    vector<string> strhelp = GetDirectoryListing(path);
    for(vector<string>::iterator iter = strhelp.begin();iter != strhelp.end();iter++)
        if(iter->find(contains)!=string::npos)
            strReturn.push_back(*iter);
    
    return strReturn;
}
	

bool CMyTools::isNumeric(const char *val)
{
	try
	{
		ostringstream o;
		o << atof(val);
//		cout << "(" << val << ")" << std::atof(val) << " - (" << o.str() << ")" << std::atof(o.str().c_str()) << " = " << std::atof(o.str().c_str())-std::atof(val) << endl;
		if(atof(o.str().c_str())-atof(val) != 0)
			throw o.str();
		return true;
	}
	catch(string v)
	{
		throw string("converting const char* to float failed for ")+val+" -> "+v;	
	}
	catch(...)
	{
		throw string("converting const char* to float failed for ")+val;	
	}
}


string CMyTools::GetComplSequence(string& strSequence)
{
	try
    {
        int intPos = 0;
    	size_t intSize = strSequence.length();
    	char* buffer=(char*)malloc(intSize+1);
        if(!buffer)
            throw(string("can't allocate enough memory."));
        buffer[intSize]=0;
        while(--intSize != -1)
        {
            buffer[intPos++]=GetComplNuc(strSequence.at(intSize));
        }        
       	string strReturn(buffer);
        free(buffer);
        return strReturn;
    }
    catch(string strError)
    {
        throw(string("error in CMyTools::GetSeqFromFile, ")+strError); 
    }  
    catch(...)  
    {
        throw("unknown error in CMyTools::GetSeqFromFile\n");                 
    }
	
}

vector<unsigned int> CMyTools::FindApproxSubs(unsigned int k, string & P, string & T)
{
  // Aufgabe: Finde alle Positionen in T, an denen ein Substring von T mit P 
  // ann�hernd �bereinstimmt (mit Abstand <= k), und gebe diese in Out aus.
	
  unsigned int n = (unsigned int )T.length();
  unsigned int m = (unsigned int )P.length();
  unsigned int *C = new unsigned int[m];
  unsigned int last = k + 1;
  unsigned int pC,nC, i, pos;
	
  vector<unsigned int>  Out;

  for(i = 0; i < m; i++) { C[i] = i + 1; }

  // Searching
  for(pos = 0; pos < n; pos++) 
  {
		pC = 0;
		nC = 0;
		//for(i = 0; i < last; i++) 
		for(i = 0; i <  last; i++) 
		{
			if(P[i] == T[pos]) 
			{
				nC = pC;
			} 
			else 
			{
				if(pC < nC) 
					nC = pC;
				if(C[i] < nC) 
					nC = C[i];
				nC++;
			}
			pC = C[i];
			C[i] = nC;
		}
		while(C[last - 1] > k) 
			last--;
		if(last == m) 
		  Out.push_back(pos + 1);
		else
			last++; 
  }
  delete C;
  return Out;
}


char CMyTools::GetComplNuc(const char& source)
{
	switch(source)
	{
	case 'A':
		return 'T';
	case 'a':
		return 't';
	case 'C':
		return 'G';
	case 'c':
		return 'g';
	case 'G':
		return 'C';
	case 'g':
		return 'c';
	case 'T':
	case 'U':
		return 'A';
	case 't':
	case 'u':
		return 'a';
	case 'M':
		return 'K';
	case 'm':
		return 'k';
	case 'R':
		return 'Y';
	case 'r':
		return 'y';
	case 'Y':
		return 'R';
	case 'y':
		return 'r';
	case 'K':
		return 'M';
	case 'k':
		return 'm';
	case 'V':
		return 'B';
	case 'v':
		return 'b';
	case 'H':
		return 'D';
	case 'h':
		return 'd';
	case 'D':
		return 'H';
	case 'd':
		return 'h';
	case 'B':
		return 'V';
	case 'b':
		return 'v';
	case '(':
         return ')';
    case ')':
         return '(';
	case '[':
         return ']';
    case ']':
         return '[';
	}
	return source;
}

char CMyTools::decodeBaseChange(char base, char cc)
{
	try
	{
		switch(base)
		{
			case 'a':
			case 'A':
				switch(cc)
				{
					case '0':return 'A';break;
					case '1':return 'C';break;
					case '2':return 'G';break;
					case '3':return 'T';break;
					default:return 'N';
				}
				break;
			case 'c':
			case 'C':
				switch(cc)
				{
					case '0':return 'C';break;
					case '1':return 'A';break;
					case '2':return 'T';break;
					case '3':return 'G';break;
					default:return 'N';
				}
				break;
			case 'g':
			case 'G':
				switch(cc)
				{
					case '0':return 'G';break;
					case '1':return 'T';break;
					case '2':return 'A';break;
					case '3':return 'C';break;
					default:return 'N';
				}
				break;
			case 't':
			case 'T':
				switch(cc)
				{
					case '0':return 'T';break;
					case '1':return 'G';break;
					case '2':return 'C';break;
					case '3':return 'A';break;
					default:return 'N';
				}
				break;
			default:
				return 'N';
		}		
	}
	catch(const char* err)
	{
		throw(err);
	}
	catch(string& err)
	{
		throw(err);
	}
	catch (exception& err)
        {
      		throw(err);
    	}
    	catch(...)
	{
		throw(1);
	}	
}


string CMyTools::getBasespace(string cs,bool keep_lead_base)
{
	try
	{
		string	bs;
		char	lead_base;
		string::iterator iter = cs.begin();
                if(keep_lead_base)bs=*iter;
		if(iter!=cs.end())
		{
			lead_base=*iter;
			for(iter++;iter!=cs.end();iter++)
			{
				lead_base=decodeBaseChange(lead_base,*iter);
				bs+=lead_base;
			}
		}
		return bs;		
	}
	catch(const char* err)
	{
		throw(err);
	}
	catch(string& err)
	{
		throw(err);
	}
	catch (exception& err)
    {
  		throw(err);
	}
	catch(...)
	{
		throw(1);
	}			
}

string CMyTools::getBasespaceRev(string cs,bool keep_lead_base)
{
	try
	{
		string	bs;
                char	lead_base;
		string::reverse_iterator iter = cs.rbegin();
                if(keep_lead_base)bs=*iter;
		if(iter!=cs.rend())
		{
			lead_base=*iter;
			for(iter++;iter!=cs.rend();iter++)
			{
				lead_base=decodeBaseChange(lead_base,*iter);
				bs+=lead_base;
			}
		}
		return bs;		
	}
	catch(const char* err)
	{
		throw(err);
	}
	catch(string& err)
	{
		throw(err);
	}
	catch (exception& err)
    {
  		throw(err);
	}
	catch(...)
	{
		throw(1);
	}			
}

template<class T> void CMyTools::GetMem(T*& OldMem, int Elems)
{
    typedef int cntr; // Type of element cntr
    const int CountSize = sizeof(cntr); // And size
    const int TypeSize = sizeof(T);
    if( Elems == 0) 
    {
        free( &(((cntr*)OldMem)[-1]) );
        return;
    }

    T* p = OldMem;
    cntr OldCount = 0;
    if(p) 
    { // Previously allocated memory
        cntr* tmp = reinterpret_cast<cntr*>(p);
        p = reinterpret_cast<T*>(--tmp);
        OldCount = *(cntr*)p; // Previous # Elems
    }

    T* m = (T*)realloc(p, Elems * TypeSize + CountSize);
    assert(m != 0);
    *((cntr*)m) = Elems; // Keep track of count
    const cntr Increment = Elems - OldCount;
    if( Increment > 0) 
    {
        // Starting address of data:
        long StartAdr = (long)&(m[OldCount]);
        StartAdr += CountSize;
        // Zero the additional new memory:
        memset((void*)StartAdr, 0, Increment * TypeSize);
    }
    // Return the address beyond the count:
    OldMem = (T*)&(((cntr*)m)[1]);
}

string CMyTools::dtoa(long value, int base) 
{
    string strReturn;
    strReturn.reserve( MYTOOLS_H_MAX_ITOA_DIGITS_SIZE ); 
    const   char  translator[] = "0123456789abcdef";
    
    bool    negative = value < 0;
    value   = abs(value);
    if (base < 2 || base > 16) 
        return strReturn;

     while (value)
     {
        strReturn += translator[value % base];
        value /= base;
     }

    if (negative) 
        strReturn += '-';
    
    reverse( strReturn.begin(), strReturn.end() );
    return strReturn;
}

