/* 
 * File:   main.cpp
 * Author: mwittig
 *
 * Created on March 22, 2013, 1:09 PM
 */


#include <cstdio>
#include <cstdlib>
#include <string>
#include <map>
#include <fstream>
#include <sstream>
#include <iostream>
#include <iomanip>
#include <set>
#include <vector>
#include <dirent.h>

#include "CMultiFasta.h"
#include "CMWException.h"
#include "MyTools.h"
/*
 * 
 */

using namespace std;

vector<string> Tokenize(const string& str, const string& delimiters = "\t")
{
    vector<string> tokens;
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        if(str.length() > pos && delimiters.find(str[pos]) != string::npos )
            lastPos=pos+1;
        else
            lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
    return tokens;
}


int main(int argc, char** argv) {

    try
    {
        // --Ref /home/mwittig/data/HLA/annotation/original/v3.9/cDNA_v3.9_exon_wise.fasta --fastq /home/mwittig/data/HLA/newAGI/B0708_hla_gen_nuc_390_fragment.fastq --read_length 100 --min_alignment 30
        int         arg_read_length = 100;
        int         arg_min_align   = 30;
        string      arg_ref_file    = "reference.fa";
        string      arg_fastq_file  = "reads.fastq";
        string      arg_on_target_file = "./on-target.fastq";
        string      arg_off_target_file  = "./off-target.fastq";

        if(argc != 13)
        { // Check the value of argc. If not enough parameters have been passed, inform user and exit.
            cout << "Usage is:" << endl << argv[0] << " --Ref <infile> --fastq <infile> --read_length <integer> --min_alignment <integer>" << endl;
            cout      << "         " << "--Ref          : Reference as fasta (exon wise for cDNA!!!)" << endl;
            cout      << "         " << "--fastq        : Reads to align" << endl;
            cout      << "         " << "--read_length  : length of reads." << endl;
            cout      << "         " << "--min_alignment: Minimum alignment for soft clipped reads. Softclipping only at exon boundaries " << endl;
            cout      << "         " << "--on-target    : store on target reads in this file" << endl;
            cout      << "         " << "--off-target   : store off target reads in this file " << endl;
            exit(EXIT_FAILURE);
        }
        else
        {
            cerr << "  starting " << argv[0] << endl << "  arguments:" << endl;
            for (int i = 1; i < argc-1; i+=2)
            {
                if(string(argv[i]).compare("--Ref")==0)
                    arg_ref_file = argv[i+1];
                else if(string(argv[i]).compare("--fastq")==0)
                    arg_fastq_file = argv[i+1];
                else if(string(argv[i]).compare("--on-target")==0)
                    arg_on_target_file = argv[i+1];
                else if(string(argv[i]).compare("--off-target")==0)
                    arg_off_target_file = argv[i+1];
                else if(string(argv[i]).compare("--read_length")==0)
                    arg_read_length = atoi(argv[i+1]);
                else if(string(argv[i]).compare("--min_alignment")==0)
                    arg_min_align = atoi(argv[i+1]);
                else
                    throw(string("Unexpected argument ")+argv[i]+", check spelling !!!");
                cerr << setw(11) << right << argv[i] << ' ' << argv[i+1] << endl;
            }
         }

        set<string> full_length_index;
        set<string> softclipped_index;

        CMultiFasta fasta(arg_ref_file);
        for(CMultiFasta::iterator iter = fasta.begin(); iter != fasta.end(); iter++)
        {
            if(iter->second.size() < arg_min_align)
                continue;
            //cout << iter->second.size() << endl;
            softclipped_index.insert(iter->second.substr(0,arg_min_align));
            softclipped_index.insert(iter->second.substr(iter->second.size()-arg_min_align,arg_min_align));
            if(iter->second.size() < arg_read_length)
                continue;
            for(size_t i = 0; i <= iter->second.size()-arg_read_length; i++)
                full_length_index.insert(iter->second.substr(i,arg_read_length));

        }
        cerr << "full length index: " << full_length_index.size() << endl
             << arg_min_align << "bp index: " << softclipped_index.size()  << endl;

        ifstream in(arg_fastq_file.c_str());
        if(!in)
            throw(string("Can't open ")+arg_fastq_file);
        ofstream outOn(arg_on_target_file.c_str());
        if(!outOn)
            throw(string("Can't open ")+arg_on_target_file);
        ofstream outOff(arg_off_target_file.c_str());
        if(!outOff)
            throw(string("Can't open ")+arg_off_target_file);
        size_t read_counter=0;
        size_t mappable_counter=0;
        size_t to_short=0;
        string strLine1;
        string strLine2;
        string strLine3;
        string strLine4;
        while(getline(in,strLine1) && getline(in,strLine2) && getline(in,strLine3) && getline(in,strLine4))
        {
            string seq_line = strLine2;
            string qual_line = strLine4;
            strLine2=strLine2.substr(0,arg_read_length);
            strLine4=strLine4.substr(0,arg_read_length);
            read_counter++;
            if(strLine2.size() < arg_read_length)
            {
                cerr << "You set read length is " << arg_read_length << " but read " << strLine1 << " is " << strLine2.size() << endl;
                to_short++;
                continue;
                // throw("read length mismatch error");
            }
            bool is_mappable = false;
            string compl_seq = CMyTools::GetComplSequence(strLine2);
            if(full_length_index.find(strLine2) != full_length_index.end() || full_length_index.find(compl_seq) != full_length_index.end())
                is_mappable = true;

            for(size_t i = 0; i <= arg_read_length-arg_min_align && !is_mappable;i++)
                if(softclipped_index.find(strLine2.substr(i,arg_min_align)) != softclipped_index.end() || softclipped_index.find(compl_seq.substr(i,arg_min_align)) != softclipped_index.end())
                    is_mappable = true;

            if(is_mappable)
            {
                //cout << Tokenize(strLine1," \t")[0] << endl << strLine2 << endl << strLine3 << endl << strLine4 << endl;
                outOn << strLine1 << endl << seq_line << endl << strLine3 << endl << qual_line << endl;
                mappable_counter++;
            }
            else
                outOff << strLine1 << endl << seq_line << endl << strLine3 << endl << qual_line << endl;
        }
        in.close();
        outOn.close();
        outOff.close();
        
        cerr << mappable_counter << '/' << read_counter << " reads mapped" << endl;
        cerr << to_short << " reads are shorter than expected" << endl;
        return (EXIT_SUCCESS);
    }
    catch(const string& err)
    {
        cerr << err << endl;
    }
    catch(const char* err)
    {
        cerr << err << endl;
    }
    catch(CMWException err)
    {
        cerr << err.what() << endl;
    }
    catch(...)
    {
        cerr << "Unexpected error ..." << endl;
    }
    return EXIT_FAILURE;
}

